package com.mortals.common.cluster.dispatcher;

import com.alibaba.fastjson.JSON;
import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.mqtt.broker.common.helper.RejectHandler;
import com.mortals.mqtt.broker.common.helper.ThreadFactoryImpl;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.model.Message;
import com.mortals.mqtt.broker.common.model.MessageHeader;
import com.mortals.mqtt.broker.common.model.Subscription;
import com.mortals.mqtt.broker.processor.MemoryMessageHandler;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.bootstrap.util.MessageUtil;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import io.netty.handler.codec.mqtt.MqttPublishMessage;
import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.*;

/**
 * 默认的消息分发实现类
 */
public class DefaultDispatcherInnerMessage extends MemoryMessageHandler implements InnerMessageDispatcher {

    private static final Logger log = MqttLogger.messageTraceLog;
    private boolean stoped = false;
    private static final BlockingQueue<Message> messageQueue = new LinkedBlockingQueue<>(100000);
    //管理端下发控制消息，与客服端发布消息队列分离
    private static final BlockingQueue<Message> messageCtrQueue = new LinkedBlockingQueue<>(100000);

    private ThreadPoolExecutor pollThread;
    private int pollThreadNum;
    private SubscriptionMatcher subscriptionMatcher;
    private SessionStore sessionStore;
    private ClusterEventHandler clusterEventHandler;


    public DefaultDispatcherInnerMessage(MqttBrokerService mqttBrokerService) {
        super(mqttBrokerService);
        this.pollThreadNum = mqttBrokerService.getBrokerConfig().getPollThreadNum();
        this.subscriptionMatcher = mqttBrokerService.getSubscriptionMatcher();
        this.sessionStore = mqttBrokerService.getSessionStore();
        this.clusterEventHandler = mqttBrokerService.getClusterEventHandler();
    }

    @Override
    public void start() {
        //创建线程池
        this.pollThread = new ThreadPoolExecutor(pollThreadNum,
                pollThreadNum,
                60 * 1000,
                TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<>(100000),
                new ThreadFactoryImpl("pollMessage2Subscriber"),
                new RejectHandler("pollMessage", 100000));
       //启动一个线程无限循环获取消息队列中的消息，一次消息队列为32条 每个一秒取一次
        new Thread(() -> {
            //队列获取超时时间
            int waitTime = 1000;
            while (!stoped) {
                try {
                    List<Message> messageList = new ArrayList(32);
                    Message message;
                    for (int i = 0; i < 32; i++) {
                        if (i == 0) {
                            message = messageQueue.poll(waitTime, TimeUnit.MILLISECONDS);
                        } else {
                            message = messageQueue.poll();
                        }
                        if (Objects.nonNull(message)) {
                            messageList.add(message);
                        } else {
                            break;
                        }
                    }
                    if (messageList.size() > 0) {
                        LogUtil.debug(log, "poll messageQueue messages."+ JSON.toJSONString(messageList));
                        //异步消息分发
                        AsyncDispatcher dispatcher = new AsyncDispatcher(messageList);
                        pollThread.submit(dispatcher).get();
                    }
                } catch (InterruptedException e) {
                    LogUtil.warn(log, "poll message wrong.");
                } catch (ExecutionException e) {
                    LogUtil.warn(log, "AsyncDispatcher get() wrong.");
                }
            }
        }).start();

        new Thread(() -> {
            int waitTime = 1000;
            while (!stoped) {
                try {
                    List<Message> messageList = new ArrayList(32);
                    Message message;
                    for (int i = 0; i < 32; i++) {
                        if (i == 0) {
                            message = messageCtrQueue.poll(waitTime, TimeUnit.MILLISECONDS);
                        } else {
                            message = messageCtrQueue.poll();
                        }
                        if (Objects.nonNull(message)) {
                            messageList.add(message);
                        } else {
                            break;
                        }
                    }
                    if (messageList.size() > 0) {
                        LogUtil.debug(log, "poll messageCtrQueue messages."+ JSON.toJSONString(messageList));
                        //异步消息分发
                        AsyncDispatcher dispatcher = new AsyncDispatcher(messageList);
                        pollThread.submit(dispatcher).get();
                    }
                } catch (InterruptedException e) {
                    LogUtil.warn(log, "poll message wrong.");
                } catch (ExecutionException e) {
                    LogUtil.warn(log, "AsyncDispatcher get() wrong.");
                }
            }
        }).start();
    }

    @Override
    public boolean appendMessage(Message message) {
        boolean isNotFull = messageQueue.offer(message);
        if (!isNotFull) {
            LogUtil.warn(log, "[PubMessage] -> the buffer messageQueue queue is full");
        }
        return isNotFull;
    }

    @Override
    public boolean appendControlMessage(Message message) {
        boolean isNotFull = messageCtrQueue.offer(message);
        if (!isNotFull) {
            LogUtil.warn(log, "[PubMessage] -> the buffer messageCtrQueue queue is full");
        }
        return isNotFull;
    }

    @Override
    public void shutdown() {
        this.stoped = true;
        this.pollThread.shutdown();
    }

    class AsyncDispatcher implements Runnable {

        private List<Message> messages;

        AsyncDispatcher(List<Message> messages) {
            this.messages = messages;
        }

        @Override
        public void run() {
            if (Objects.nonNull(messages)) {
                try {
                    for (Message message : messages) {
                        LogUtil.debug(log, "[WRITE MESSAGE] ->message:{}",JSON.toJSONString(message));

                        //通过message获取订阅者，遍历每个订阅者
                        Set<Subscription> subscriptions = subscriptionMatcher.match((String) message.getHeader(MessageHeader.TOPIC), message.getClientId());
                        for (Subscription subscription : subscriptions) {
                            String clientId = subscription.getClientId();
                            //查看当前连接管理中是否包含clientId
                            if (ConnectManager.getInstance().containClient(clientId)) {
                                //获取连接中的sessiong qos 根据qos不同指定不同的发送策略
                                ClientSession clientSession = ConnectManager.getInstance().getClient(clientId);
                                int qos = MessageUtil.getMinQos((int) message.getHeader(MessageHeader.QOS), subscription.getQos());
                                int messageId = clientSession.generateMessageId();
                                //构造消息头信息
                                message.putHeader(MessageHeader.QOS, qos);
                                message.setMsgId(messageId);
                                if (qos > 0) {
                                    //1|2 需要缓冲消息
                                    cacheOutflowMsg(clientId, message);
                                }
                                //给所有订阅者发布消息
                                MqttPublishMessage publishMessage = MessageUtil.getPubMessage(message, false);
                                LogUtil.debug(log, "[WRITE MESSAGE NOW] ->message LENS:{}",publishMessage.toString());
                                clientSession.getCtx().writeAndFlush(publishMessage);
                            } else {
                                //连接未找到，持久化离线消息
                                LogUtil.debug(log, "[WRITE OFFLINE MESSAGE] -> clientId:{},message:{}", clientId,JSON.toJSONString(message));

                                sessionStore.storeOfflineMsg(clientId, message);
                            }
                        }
                    }
                } catch (Exception ex) {
                    log.error("异常",ex);
                    LogUtil.warn(log, "Dispatcher message failure,cause={}", ex);
                }
            }
        }
    }
}
