package com.mortals.common.cluster.dispatcher;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.TypeReference;
import com.mortals.mqtt.broker.common.model.Subscription;
import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.model.Message;
import com.mortals.common.cluster.dispatcher.event.Event;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.store.SubscriptionStore;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import org.slf4j.Logger;

import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * 集群事件处理
 */
public class EventConsumeHandler {

    private static final Logger log = MqttLogger.eventLog;

    private InnerMessageDispatcher innerMessageDispatcher;
    private ClusterEventHandler clusterEventHandler;
    private AtomicBoolean pollStoped = new AtomicBoolean(false);
    private int maxPollNum;
    private int pollWaitInterval;
    private String currentIp;
    private SessionStore sessionStore;
    private SubscriptionMatcher subscriptionMatcher;
    private SubscriptionStore subscriptionStore;

    public EventConsumeHandler(MqttBrokerService mqttBrokerService) {
        this.innerMessageDispatcher = mqttBrokerService.getInnerMessageDispatcher();
        this.clusterEventHandler = mqttBrokerService.getClusterEventHandler();
        this.maxPollNum = mqttBrokerService.getBrokerConfig().getMaxPollEventNum();
        this.pollWaitInterval = mqttBrokerService.getBrokerConfig().getPollWaitInterval();
        this.currentIp = mqttBrokerService.getCurrentIp();
        this.sessionStore = mqttBrokerService.getSessionStore();
        this.subscriptionMatcher = mqttBrokerService.getSubscriptionMatcher();
        this.subscriptionStore= mqttBrokerService.getSubscriptionStore();
    }

    // 集群方式1: consume event from cluster
    public void consumeEvent(Event event) {
        //获取事件的类型
        switch (event.getEventCode()) {
            case 1:
                clearClientSession(event);
                break;
            case 2:
                dispatcherMessage(event);
                break;
            case 3:
                dispatcherMessage(event);
                break;
            case 4:
                dispatcherControlMessage(event);
                break;
            case 5:
                dispatcherSubsMessage(event);
                break;
            case 6:
                resolveSysMessage(event);
                break;
            default:
                LogUtil.warn(log, "[EventConsumeHandler] consume event is not supported,event:{}", event);
        }
    }

    // 集群方式2: poll event from cluster
    private List<Event> pollEvent() {
        return clusterEventHandler.pollEvent(maxPollNum);
    }

    public void start() {
        clusterEventHandler.setEventConsumeHandler(this);

        new Thread(() -> {
            while (!pollStoped.get()) {
                try {
                    List<Event> eventList = pollEvent();
                    if (!MixAll.isEmpty(eventList)) {
                        for (Event event : eventList) {
                            consumeEvent(event);
                        }
                    }
                    if (MixAll.isEmpty(eventList) || eventList.size() < 5) {
                        Thread.sleep(pollWaitInterval);
                    }
                } catch (Exception e) {
                    LogUtil.warn(log, "Poll event from cluster error.", e);
                }
            }
        }).start();
    }

    public void shutdown() {

    }

    void dispatcherMessage(Event event) {
        Message message = JSONObject.parseObject(event.getBody(), Message.class);
        this.innerMessageDispatcher.appendMessage(message);
    }

    void dispatcherControlMessage(Event event) {
        Message message = JSONObject.parseObject(event.getBody(), Message.class);
        this.innerMessageDispatcher.appendMessage(message);
    }

    void dispatcherSubsMessage(Event event) {
        //消费订阅事件，同步更新树
        //Set<Subscription> subscriptions = JSON.parseObject(event.getBody(),new TypeReference<Set<Subscription>>(){});
        Message message = JSONObject.parseObject(event.getBody(), Message.class);
        //更新当前订阅树
        Set<Subscription> subscriptionSet = JSON.parseObject(MixAll.convertByteS2Str(message.getPayload()), new TypeReference<Set<Subscription>>() {
        });

        if (!MixAll.isEmpty(subscriptionSet)) {
            subscriptionSet.forEach(item -> {
                subscriptionMatcher.subscribe(item);
            });
        }

    }


   public void clearClientSession(Event event) {
        //忽略当前节点
        if (currentIp.equals(event.getFromIp())) {
            LogUtil.debug(log, "Event from current node,ignore the event,fromIp:{}", event.getFromIp());
            return;
        }
        String clientId = event.getBody();
        if (!ConnectManager.getInstance().containClient(clientId)) {
            return;
        }
        ClientSession clientSession = ConnectManager.getInstance().getClient(clientId);
        clientSession.getCtx().close();
        //获取订阅信息，并删除
        Set<Subscription> subscriptionSet = subscriptionStore.getSubscriptions(clientId);
        if (!MixAll.isEmpty(subscriptionSet)) {
            subscriptionSet.forEach(item -> {
                subscriptionMatcher.unSubscribe(item.getTopic(), clientId);
            });
        }
    }

    public void resolveSysMessage(Event event) {
        //处理管理端来的系统消息

        System.out.println("system :"+event.getBody());

    }

}
