package com.mortals.common.queue;

import com.alibaba.fastjson.JSONObject;
import com.mortals.mqtt.broker.common.helper.DateUtils;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.model.Message;
import com.mortals.mqtt.broker.common.model.MessageHeader;
import com.mortals.mqtt.broker.common.model.MsgTypeEnum;
import com.mortals.common.cluster.dispatcher.ClusterEventHandler;
import com.mortals.common.cluster.dispatcher.event.Event;
import com.mortals.common.cluster.dispatcher.event.EventCode;
import lombok.Getter;
import org.slf4j.Logger;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * @author: finegirl
 * @date: 2021/5/20 13:49
 * @description: 消费消息服务
 **/
public class ConsumerService {
    private static final Logger log = MqttLogger.eventLog;
    private long pollDuration;
    protected volatile ExecutorService consumersExecutor;
    @Getter
    private TbQueueConsumer<TbQueueMsg> mainConsumer;
    private ClusterEventHandler clusterEventHandler;
    private String currentIp;

    protected volatile boolean stopped = false;


    public void init(TbQueueConsumer<TbQueueMsg> mainConsumer) {
        this.consumersExecutor = Executors.newCachedThreadPool();
        this.mainConsumer = mainConsumer;
        launchMainConsumers();
        this.mainConsumer.subscribe();

    }

    public ConsumerService(long pollDuration, ClusterEventHandler clusterEventHandler, String currentIp) {
        this.pollDuration = pollDuration;
        this.clusterEventHandler = clusterEventHandler;
        this.currentIp = currentIp;
    }

    protected void launchMainConsumers() {
        consumersExecutor.submit(() -> {
            while (!stopped) {
                try {

                    List<TbQueueMsg> poll = mainConsumer.poll(pollDuration);

                    List<TbQueueMsg> msgs = poll;
                    if (msgs.isEmpty()) {
                        continue;
                    }

                    for (TbQueueMsg item:msgs){
                        Message innerMsg = new Message();
                        TbQueueMsgHeaders headMap = item.getHeaders();
                        int messageProtocl = headMap.get(MessageHeader.MESSAGEPROTOCOL)[0];
                        Map<String, Object> headers = new HashMap<>();
                        innerMsg.setPayload(item.getData());
                        headers.put(MessageHeader.MESSAGEPROTOCOL, messageProtocl);
                        innerMsg.setStoreTime(DateUtils.getCurrDate().getTime());
                        if(messageProtocl== MsgTypeEnum.MSG_SYSTEM.getValue()){
                            innerMsg.setHeaders(headers);
                            sendSysMessage2Cluster(innerMsg);
                        }else{
                            String clientId = MixAll.convertByteS2Str(headMap.get(MessageHeader.CLIENTID));
                            int messageType = headMap.get(MessageHeader.MESSAGETYPE)[0];
                            String topic = MixAll.convertByteS2Str(headMap.get(MessageHeader.TOPIC));
                            int qos = headMap.get(MessageHeader.QOS)[0];

                            innerMsg.setClientId(clientId);
                            innerMsg.setType(Message.Type.valueOf(messageType));
                            headers.put(MessageHeader.TOPIC, topic);
                            headers.put(MessageHeader.QOS, qos);
                            headers.put(MessageHeader.RETAIN, false);
                            headers.put(MessageHeader.DUP, false);
                            innerMsg.setHeaders(headers);
                            sendContrlMessage2Cluster(innerMsg);
                        }
                    }
                } catch (Exception e) {
                    LogUtil.error(log,"comsumer error",e);
                }
            }
            LogUtil.info(log,"Queue Consumer stopped.");
        });
    }


    public void destroy() {
        if (!stopped) {
            stopMainConsumers();
        }
        stopped = true;
        if (consumersExecutor != null) {
            consumersExecutor.shutdownNow();
        }

    }

    protected void stopMainConsumers() {
        if (mainConsumer != null) {
            mainConsumer.unsubscribe();
        }
    }

    /**
     * 向集群分发消息:第一阶段
     */
    private void sendContrlMessage2Cluster(Message message) {
        Event event = new Event(EventCode.DISPATCHER_CONTROL_CLIENT_MESSAGE.getCode(), JSONObject.toJSONString(message), System.currentTimeMillis(), currentIp);
        this.clusterEventHandler.sendEvent(event);
    }

    private void sendSysMessage2Cluster(Message message) {
        Event event = new Event(EventCode.DISPATCHER_SYS_MESSAGE.getCode(), JSONObject.toJSONString(message), System.currentTimeMillis(), currentIp);
        this.clusterEventHandler.sendEvent(event);
    }

}
