/**
 * Copyright © 2016-2021 The Thingsboard Authors
 * <p>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.mortals.common.queue.kafka;

import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.common.queue.TbQueueCallback;
import com.mortals.common.queue.TbQueueMsg;
import com.mortals.common.queue.TbQueueProducer;
import com.mortals.common.queue.TopicPartitionInfo;
import lombok.Builder;
import lombok.Data;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.header.Header;
import org.apache.kafka.common.header.internals.RecordHeader;
import org.slf4j.Logger;

import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

@Data
public class TbKafkaProducerTemplate<T extends TbQueueMsg> implements TbQueueProducer<T> {
    private static final Logger log = MqttLogger.messageTraceLog;
    private KafkaProducer<String, byte[]> producer;

    private String defaultTopic;
    private Set<TopicPartitionInfo> topics;

    @Builder
    private TbKafkaProducerTemplate(String defaultTopic,  Properties props) {
        this.producer = new KafkaProducer<>(props);
        this.defaultTopic = defaultTopic;
        topics = ConcurrentHashMap.newKeySet();
    }

    @Override
    public void init() {
    }

    @Override
    public void send(TopicPartitionInfo tpi, T msg, TbQueueCallback callback) {
        String key = msg.getKey().toString();
        byte[] data = msg.getData();
        ProducerRecord<String, byte[]> record;
        if(tpi.getTopic()==null){
            tpi.setTopic(this.defaultTopic);
        }
        Iterable<Header> headers = msg.getHeaders().getData().entrySet().stream().map(e -> new RecordHeader(e.getKey(), e.getValue())).collect(Collectors.toList());
        record = new ProducerRecord<>(tpi.getTopic(), null, key, data, headers);
        producer.send(record, (metadata, exception) -> {
            if (exception == null) {
                if (callback != null) {
                    callback.onSuccess(new KafkaTbQueueMsgMetadata(metadata));
                }
            } else {
                if (callback != null) {
                    callback.onFailure(exception);
                } else {
                    log.warn("Producer template failure: {}", exception.getMessage(), exception);
                }
            }
        });

    }


    @Override
    public void stop() {
        if (producer != null) {
            producer.close();
        }
    }


}
