package com.mortals.mqtt.broker.bootstrap.netty;

import com.mortals.mqtt.broker.bootstrap.NettyService;
import com.mortals.mqtt.broker.bootstrap.netty.codec.ByteBuf2WebSocketEncoder;
import com.mortals.mqtt.broker.bootstrap.netty.codec.WebSocket2ByteBufDecoder;
import com.mortals.mqtt.broker.common.config.BrokerConfig;
import com.mortals.mqtt.broker.common.config.NettyConfig;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.helper.Pair;
import com.mortals.mqtt.broker.common.helper.ThreadFactoryImpl;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.dropping.MessageDroppedService;
import com.mortals.mqtt.broker.metrics.BytesMetricsCollector;
import com.mortals.mqtt.broker.metrics.MessageMetricsCollector;
import com.mortals.mqtt.broker.metrics.MessageMetricsHandler;
import com.mortals.mqtt.broker.metrics.common.GlobalMQTTMessageCounter;
import com.mortals.mqtt.broker.metrics.common.GlobalTrafficCounter;
import com.mortals.mqtt.broker.metrics.common.MetricsHolder;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.buffer.PooledByteBufAllocator;
import io.netty.channel.*;
import io.netty.channel.epoll.EpollEventLoopGroup;
import io.netty.channel.epoll.EpollServerSocketChannel;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.codec.http.HttpContentCompressor;
import io.netty.handler.codec.http.HttpObjectAggregator;
import io.netty.handler.codec.http.HttpServerCodec;
import io.netty.handler.codec.http.websocketx.WebSocketServerProtocolHandler;
import io.netty.handler.codec.mqtt.MqttDecoder;
import io.netty.handler.codec.mqtt.MqttEncoder;
import io.netty.handler.codec.mqtt.MqttMessageType;
import io.netty.handler.timeout.IdleStateHandler;
import org.slf4j.Logger;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;


public class NettyNettyServer implements NettyService {
    int coreThreadNum = Runtime.getRuntime().availableProcessors();
    private static final Logger log = MqttLogger.remotingLog;
    private BytesMetricsCollector bytesMetricsCollector = new BytesMetricsCollector();
    private MessageMetricsCollector metricsCollector = new MessageMetricsCollector();
    private NettyConfig nettyConfig;
    private EventLoopGroup selectorGroup;
    private EventLoopGroup ioGroup;
    private Class<? extends ServerChannel> clazz;
    private Map<MqttMessageType, Pair<RequestProcessor, ExecutorService>> processorTable;
    private NettyEventExecutor nettyEventExecutor;
    private BrokerConfig brokerConfig;

    private Optional<? extends ChannelInboundHandler> metrics;
    private GlobalTrafficCounter trafficCounter ;
    private Optional<? extends ChannelInboundHandler> errorsCather;

    private GlobalMQTTMessageCounter globalMQTTMessageCounter;

    private MetricsHolder metricsHolder;

    private MessageDroppedService messageDroppedService;


    public NettyNettyServer(BrokerConfig brokerConfig, NettyConfig nettyConfig, ChannelEventListener listener, MetricsHolder metricsHolder,MessageDroppedService messageDroppedService) {
        this.nettyConfig = nettyConfig;
        this.processorTable = new HashMap();
        this.brokerConfig = brokerConfig;
        this.metricsHolder=metricsHolder;
        this.nettyEventExecutor = new NettyEventExecutor(listener);
        this.messageDroppedService=messageDroppedService;

        if (!nettyConfig.isUseEpoll()) {
            selectorGroup = new NioEventLoopGroup(coreThreadNum,
                    new ThreadFactoryImpl("SelectorEventGroup"));
            ioGroup = new NioEventLoopGroup(coreThreadNum * 2,
                    new ThreadFactoryImpl("IOEventGroup"));
            clazz = NioServerSocketChannel.class;
        } else {
            selectorGroup = new EpollEventLoopGroup(coreThreadNum,
                    new ThreadFactoryImpl("SelectorEventGroup"));
            ioGroup = new EpollEventLoopGroup(coreThreadNum * 2,
                    new ThreadFactoryImpl("IOEventGroup"));
            clazz = EpollServerSocketChannel.class;
        }


        boolean useFineMetrics = nettyConfig.isUseMetrics();
        if (useFineMetrics) {
            globalMQTTMessageCounter = new GlobalMQTTMessageCounter(metricsHolder,messageDroppedService);
            //DropWizardMetricsHandler metricsHandler = new DropWizardMetricsHandler();
            //metricsHandler.init(brokerConfig.getMetricRegistry());
           // this.metrics = Optional.of(metricsHandler);
            //统计流量
            trafficCounter = new GlobalTrafficCounter(brokerConfig.getMetricRegistry(), ioGroup);
            trafficCounter.init();
        } else {
            this.metrics = Optional.empty();
        }

        //是否开启实施监控错误报告
        final boolean useBugSnag = nettyConfig.getBugsnagEnable();
        if (useBugSnag) {
            BugSnagErrorsHandler bugSnagHandler = new BugSnagErrorsHandler();
            bugSnagHandler.init();
            this.errorsCather = Optional.of(bugSnagHandler);
        } else {
            this.errorsCather = Optional.empty();
        }
    }


    @Override
    public void start() {
        //Netty event executor start
        this.nettyEventExecutor.start();
        // start TCP server
        if (nettyConfig.isStartTcp()) {
            startTcpServer(false, nettyConfig.getTcpPort());
        }

        if (nettyConfig.isStartSslTcp()) {
            startTcpServer(true, nettyConfig.getSslTcpPort());
        }

        // start Websocket server
        if (nettyConfig.isStartWebsocket()) {
            startWebsocketServer(false, nettyConfig.getWebsocketPort());
        }

        if (nettyConfig.isStartSslWebsocket()) {
            startWebsocketServer(true, nettyConfig.getSslWebsocketPort());
        }
    }

    private void startWebsocketServer(boolean useSsl, Integer port) {
        ServerBootstrap bootstrap = new ServerBootstrap();
        bootstrap.group(selectorGroup, ioGroup)
                .channel(clazz)
                .option(ChannelOption.SO_BACKLOG, nettyConfig.getTcpBackLog())
                .childOption(ChannelOption.TCP_NODELAY, nettyConfig.isTcpNoDelay())
                .childOption(ChannelOption.SO_SNDBUF, nettyConfig.getTcpSndBuf())
                .option(ChannelOption.SO_RCVBUF, nettyConfig.getTcpRcvBuf())
                .option(ChannelOption.SO_REUSEADDR, nettyConfig.isTcpReuseAddr())
                .childOption(ChannelOption.SO_KEEPALIVE, nettyConfig.isTcpKeepAlive())
                .childHandler(new ChannelInitializer<SocketChannel>() {
                    @Override
                    protected void initChannel(SocketChannel socketChannel) throws Exception {
                        ChannelPipeline pipeline = socketChannel.pipeline();
                        if (useSsl) {
                            pipeline.addLast("ssl", NettySslHandler.getSslHandler(
                                    socketChannel,
                                    nettyConfig.isUseClientCA(),
                                    nettyConfig.getSslKeyStoreType(),
                                    brokerConfig.getMqttHome() + nettyConfig.getSslKeyFilePath(),
                                    nettyConfig.getSslManagerPwd(),
                                    nettyConfig.getSslStorePwd()
                            ));
                        }
                        pipeline.addLast("idleStateHandler", new IdleStateHandler(0, 0, 60))
                                .addLast("httpCodec", new HttpServerCodec())
                                .addLast("aggregator", new HttpObjectAggregator(65535))
                                .addLast("compressor ", new HttpContentCompressor())
                                .addLast("webSocketHandler", new WebSocketServerProtocolHandler("/mqtt", MixAll.MQTT_VERSION_SUPPORT, false, 65536))
                                .addLast("webSocket2ByteBufDecoder", new WebSocket2ByteBufDecoder())
                                .addLast("byteBuf2WebSocketEncoder", new ByteBuf2WebSocketEncoder())
                                .addLast("mqttDecoder", new MqttDecoder(nettyConfig.getMaxMsgSize()))
                                .addLast("mqttEncoder", MqttEncoder.INSTANCE)
                                .addLast("nettyConnectionManager", new NettyConnectHandler(
                                        nettyEventExecutor))
                                .addLast("nettyMqttHandler", new NettyMqttHandler(processorTable));
                    }
                });
        if (nettyConfig.isPooledByteBufAllocatorEnable()) {
            bootstrap.childOption(ChannelOption.ALLOCATOR, PooledByteBufAllocator.DEFAULT);
        }
        try {
            ChannelFuture future = bootstrap.bind(port).sync();
            LogUtil.info(log, "Start webSocket server {}  success,port = {}", useSsl ? "with ssl" : "", port);
        } catch (InterruptedException ex) {
            LogUtil.error(log, "Start webSocket server {} failure.cause={}", useSsl ? "with ssl" : "", ex);
        }
    }

    private void startTcpServer(boolean useSsl, Integer port) {
        ServerBootstrap bootstrap = new ServerBootstrap();
        bootstrap.group(selectorGroup, ioGroup)
                .channel(clazz)
                .option(ChannelOption.SO_BACKLOG, nettyConfig.getTcpBackLog())
                .childOption(ChannelOption.TCP_NODELAY, nettyConfig.isTcpNoDelay())
                .childOption(ChannelOption.SO_SNDBUF, nettyConfig.getTcpSndBuf())
                .option(ChannelOption.SO_RCVBUF, nettyConfig.getTcpRcvBuf())
                .option(ChannelOption.SO_REUSEADDR, nettyConfig.isTcpReuseAddr())
                .childOption(ChannelOption.SO_KEEPALIVE, nettyConfig.isTcpKeepAlive())
                .childHandler(new ChannelInitializer<SocketChannel>() {
                    @Override
                    protected void initChannel(SocketChannel socketChannel) throws Exception {
                        ChannelPipeline pipeline = socketChannel.pipeline();

                        if (useSsl) {
                            pipeline.addLast("ssl", NettySslHandler.getSslHandler(
                                    socketChannel,
                                    nettyConfig.isUseClientCA(),
                                    nettyConfig.getSslKeyStoreType(),
                                    brokerConfig.getMqttHome() + nettyConfig.getSslKeyFilePath(),
                                    nettyConfig.getSslManagerPwd(),
                                    nettyConfig.getSslStorePwd()
                            ));
                        }
                        //添加流量统计，单位byte
                        if(null!=trafficCounter){
                            pipeline.addFirst("global_traffic_counter", trafficCounter);
                        }

                        //解码器
                        pipeline.addFirst("idleStateHandler", new IdleStateHandler(nettyConfig.getChannelTimeout(), 0, 0));
                        if (errorsCather.isPresent()) {
                            pipeline.addLast("bugsnagCatcher", errorsCather.get());
                        }
                       // pipeline.addFirst("bytemetrics", new BytesMetricsHandler(bytesMetricsCollector));
                        pipeline.addLast("autoflush", new AutoFlushHandler(1, TimeUnit.SECONDS));
                        pipeline.addLast("mqttEncoder", MqttEncoder.INSTANCE);
                        pipeline.addLast("mqttDecoder", new MqttDecoder(nettyConfig.getMaxMsgSize()));

                       //消息管道统计
                        //pipeline.addLast("metrics", new MessageMetricsHandler(metricsCollector));

//                        if (metrics.isPresent()) {
//                            pipeline.addLast("wizardMetrics", metrics.get());
//                        }

//                        if (metricsInitializer!=null) {
//
//                           // pipeline.addLast("global_traffic_counter", trafficCounter);
//
//                           // pipeline.addLast("metricsInitializer",metricsInitializer);
//                        }

                        if(null!=globalMQTTMessageCounter){
                            pipeline.addLast("globalMQTTMessageCounter", globalMQTTMessageCounter);
                        }


                        pipeline.addLast("nettyConnectionManager", new NettyConnectHandler(
                                nettyEventExecutor));
                        pipeline.addLast("nettyMqttHandler", new NettyMqttHandler(processorTable));
                    }
                });
        if (nettyConfig.isPooledByteBufAllocatorEnable()) {
            bootstrap.childOption(ChannelOption.ALLOCATOR, PooledByteBufAllocator.DEFAULT);
        }
        try {
            ChannelFuture future = bootstrap.bind(port).sync();
            LogUtil.info(log, "Start tcp server {} success,port = {}", useSsl ? "with ssl" : "", port);
        } catch (InterruptedException ex) {
            LogUtil.error(log, "Start tcp server {} failure.cause={}", useSsl ? "with ssl" : "", ex);
        }
    }


    @Override
    public void shutdown() {
        if (selectorGroup != null) {
            selectorGroup.shutdownGracefully();
        }
        if (ioGroup != null) {
            ioGroup.shutdownGracefully();
        }
    }

    public void registerProcessor(MqttMessageType mqttType, RequestProcessor processor, ExecutorService executorService) {
        this.processorTable.put(mqttType, new Pair<>(processor, executorService));
    }

}
