package com.mortals.mqtt.broker.bootstrap.util;

import lombok.Data;

import java.math.BigDecimal;
import java.util.Date;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * @author: finegirl
 * @date: 2021/5/19 16:45
 * @description: //令牌桶实现
 **/
@Data
public class TokenBucketNum {

    private long produceTokenTimes;

    /**
     * 每秒最多请求数量
     */

    private int maxFlowRate;

    /**
     * 每秒平均请求数量
     */

    private int avgFlowRate;

    /**
     * 队列来缓存桶数量
     */

    private LinkedBlockingQueue tokenQueue;

    /**
     * 由定时任务持续生成令牌。这样的问题在于会极大的消耗系统资源，如，某接口需要分别对每个用户做访问频率限制。
     * <p>
     * 假设系统中存在6W用户，则至多需要开启6W个定时任务来维持每个桶中的令牌数，这样的开销是巨大的。
     * <p>
     * 可以做成延迟计算的形式，每次请求令牌的时候，看当前时间是否晚与下一次生成令牌的时间，计算该段时间的令牌数，
     * <p>
     * 加入令牌桶，更新数据。
     */

    //private ScheduledExecutorService scheduledExecutorService = Executors.newSingleThreadScheduledExecutor();

    /**
     * The Mutex do not use directly.
     */

    private volatile Object mutexDoNotUseDirectly = new Object();

    /**
     * The Is start.
     */

    private volatile boolean isStart = false;

    /**
     * The constant A_CHAR.
     */

    private static final Byte A_CHAR = 'a';

    /**
     * Instantiates a new Token bucket.
     */

    private TokenBucketNum() {
    }

    /**
     * New builder token bucket.
     *
     * @return the token bucket
     */

    public static TokenBucketNum newBuilder() {
        return new TokenBucketNum();

    }

    /**
     * 每秒内最大请求数量设置
     *
     * @param maxFlowRate 每秒内最大请求数量
     * @return 当前令牌同
     */

    public TokenBucketNum maxFlowRate(int maxFlowRate) {
        this.maxFlowRate = maxFlowRate;

        return this;

    }

    /**
     * 每秒平均请求数量设置
     *
     * @param avgFlowRate 每秒平均请求数量
     * @return 当前令牌同
     */

    public TokenBucketNum avgFlowRate(int avgFlowRate) {
        this.avgFlowRate = avgFlowRate;

        return this;

    }

    /**
     * 构造者模式
     *
     * @return the token bucket
     */

    public TokenBucketNum build() {
        //初始化
        init();
        //返回当前对象
        return this;

    }

    /**
     * 初始化
     */

    private void init() {
        //初始化桶队列大小
        if (maxFlowRate > 0) {
            tokenQueue = new LinkedBlockingQueue<>(maxFlowRate);
            addTokens(avgFlowRate);
            this.produceTokenTimes = new Date().getTime();

        }
       //系统启动
        isStart = true;
    }

    /**
     * 停止任务
     */

    public void stop() {
        isStart = false;
        //scheduledExecutorService.shutdown();
    }

    /**
     * 查看任务是否执行
     *
     * @return the boolean
     */

    public boolean isStarted() {
        return isStart;

    }

    /**
     * 增加令牌
     *
     * @param tokenNum the token num
     */

    private void addTokens(Integer tokenNum) {
        // 若是桶已经满了，就不再家如新的令牌
        if(tokenNum>(maxFlowRate-tokenQueue.size())){
            tokenNum=maxFlowRate-tokenQueue.size();
        }
        for (int i = 0; i < tokenNum; i++) {
            tokenQueue.offer(A_CHAR);
        }
    }

    /**
     * 获取令牌
     * true:获取到1个令牌，非阻塞
     * false:未获取到令牌，非阻塞
     *
     * @return boolean
     */

    public boolean tryAcquire() {
        //比较当前时间和令牌生成时间
        long time = new Date().getTime();
        long sub = time - produceTokenTimes;

        if (sub > 0) {
            BigDecimal avgBigDecimal = new BigDecimal(avgFlowRate);
            BigDecimal divide = avgBigDecimal.divide(new BigDecimal(1000));
            BigDecimal multiply = divide.multiply(new BigDecimal(sub));
            addTokens(multiply.intValue());
            this.produceTokenTimes = time;
            System.out.println("addTokens:"+multiply.intValue());
            // addTokens();
        }
        synchronized (mutexDoNotUseDirectly) {
            // 否存在足够的桶数量
            if (tokenQueue.size() > 0) {
                //队列不为空时返回队首值并移除,队列为空时返回null。非阻塞立即返回。
                Byte poll = (Byte) tokenQueue.poll();
                if (poll != null) {
                    //获取到令牌
                    return true;
                }
            }
        }
        //未获取到令牌
        return false;

    }

    public static void main(String[] args) throws InterruptedException {
        TokenBucketNum tokenBucket = TokenBucketNum.newBuilder().avgFlowRate(512).maxFlowRate(1024).build();
        tokenBucket.init();
        for (int i = 1; i <= 2000; i++) {

            Thread.sleep(i);

            boolean tokens = tokenBucket.tryAcquire();
            if (tokens) {
                System.out.println("token pass --- index:" + i);
            } else {
                System.out.println("token rejuect --- index" + i);
            }

        }

    }
}
