/*
 * Copyright 2019-present HiveMQ GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.mortals.mqtt.broker.dropping;


import com.mortals.mqtt.broker.bootstrap.util.TokenBucketNum;
import com.mortals.mqtt.broker.logging.EventLog;
import com.mortals.mqtt.broker.metrics.common.MetricsHolder;

import java.text.NumberFormat;
import java.util.Locale;

public class MessageDroppedServiceImpl implements MessageDroppedService {

    private static final NumberFormat FORMAT = NumberFormat.getInstance(Locale.US);
    private final MetricsHolder metricsHolder;
    private final EventLog eventLog;

    public MessageDroppedServiceImpl( MetricsHolder metricsHolder,  EventLog eventLog) {
        this.metricsHolder = metricsHolder;
        this.eventLog = eventLog;
    }


    @Override
    public void queueFull(final String clientId, final String topic, final int qos) {
        metricsHolder.getDroppedMessageCounter().inc();
        eventLog.messageDropped(clientId, topic, qos, "The client message queue is full");
    }


    @Override
    public void qos0MemoryExceeded(final String clientId, final String topic, final int qos, final long currentMemory, final long maxMemory) {
        metricsHolder.getDroppedMessageCounter().inc();
        final String reason = "The QoS 0 memory limit exceeded, size: " + FORMAT.format(currentMemory) + " bytes, max: " + FORMAT.format(maxMemory) + " bytes";
        eventLog.messageDropped(clientId, topic, qos, reason);
    }

    @Override
    public void publishExceedByBucket(String clientId, String topic, int qos, TokenBucketNum bucketNum) {
        metricsHolder.getDroppedMessageCounter().inc();
        final String reason = "The bucket limit exceeded, size: " + bucketNum.getAvgFlowRate()+ ", max: " + bucketNum.getMaxFlowRate() ;
        eventLog.messageDropped(clientId, topic, qos, reason);
    }


    @Override
    public void notWritable(final String clientId, final String topic, final int qos) {
        metricsHolder.getDroppedMessageCounter().inc();
        eventLog.messageDropped(clientId, topic, qos, "The tcp socket was not writable");
    }


    @Override
    public void extensionPrevented(final String clientId, final String topic, final int qos) {
        metricsHolder.getDroppedMessageCounter().inc();
        eventLog.messageDropped(clientId, topic, qos, "Extension prevented onward delivery");
    }

    @Override
    public void failed(final String clientId, final String topic, final int qos) {
        metricsHolder.getDroppedMessageCounter().inc();
        eventLog.messageDropped(clientId, topic, qos, "Internal error");
    }


    @Override
    public void publishMaxPacketSizeExceeded(final String clientId, final String topic, final int qos, final long maximumPacketSize, final long packetSize) {
        metricsHolder.getDroppedMessageCounter().inc();
        final String reason = "Maximum packet size exceeded, size: " + FORMAT.format(packetSize) + " bytes, max: " + FORMAT.format(maximumPacketSize) + " bytes";
        eventLog.messageDropped(clientId, topic, qos, reason);
    }

    @Override
    public void messageMaxPacketSizeExceeded(final String clientId, final String messageType, final long maximumPacketSize, final long packetSize) {
        final String reason = "Maximum packet size exceeded, size: " + FORMAT.format(packetSize) + " bytes, max: " + FORMAT.format(maximumPacketSize) + " bytes";
        eventLog.mqttMessageDropped(clientId, messageType, reason);
    }

}
