/*
 * Copyright 2019-present HiveMQ GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.mortals.mqtt.broker.info;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import oshi.SystemInfo;

import java.io.File;

public class SystemInformationImpl implements SystemInformation {

    private static final Logger log = LoggerFactory.getLogger(SystemInformationImpl.class);
    public static final String DEVELOPMENT_VERSION = "Development Snapshot";

    private File homeFolder;
    private File configFolder;
    private File logFolder;
    private File dataFolder;
    private File pluginFolder;
    private String mqVersion;
    private final long runningSince;
    private final boolean embedded;
    private final int processorCount;

    private final boolean usePathOfRunningJar;

    public SystemInformationImpl() {
        this(false);
    }

    public SystemInformationImpl(final boolean usePathOfRunningJar) {
        this(usePathOfRunningJar, false, null, null, null);
    }

    public SystemInformationImpl(
            final boolean usePathOfRunningJar,
            final boolean embedded,
            final File configFolder,
            final File dataFolder,
            final File pluginFolder) {
        this.usePathOfRunningJar = usePathOfRunningJar;
        this.embedded = embedded;
        this.configFolder = configFolder;
        this.dataFolder = dataFolder;
        this.pluginFolder = pluginFolder;


        this.runningSince = System.currentTimeMillis();
        //setHiveMQVersion();
        //setFolders();
        processorCount = getPhysicalProcessorCount();
    }

    private int getPhysicalProcessorCount() {

        final int runtimeProcessorCount = Runtime.getRuntime().availableProcessors();
        int physicalProcessorCount;
        try {
            physicalProcessorCount = new SystemInfo().getHardware().getProcessor().getPhysicalProcessorCount();
        } catch (final Exception e) {
            log.warn(
                    "No able to determine amount of physical cores, using available amount of cores reported by the JVM as fallback");
            if (log.isTraceEnabled()) {
                log.trace("Original Exception: ", e);
            }
            physicalProcessorCount = runtimeProcessorCount;
        }

        return Math.min(physicalProcessorCount, runtimeProcessorCount);
    }


    private void setMQVersion() {
        mqVersion = null;

        if (mqVersion == null || mqVersion.length() < 1) {
            mqVersion = DEVELOPMENT_VERSION;
        }

        log.info("HiveMQ version: {}", mqVersion);
    }

    public void setMqVersion(final String mqVersion) {
        this.mqVersion = mqVersion;
    }

    @Override
    public String getMQVersion() {
        return mqVersion;
    }

    @Override
    public File getHiveMQHomeFolder() {
        return homeFolder;
    }


    @Override
    public long getRunningSince() {
        return runningSince;
    }

    private File findAbsoluteAndRelative(final String fileLocation) {
        final File file = new File(fileLocation);
        if (file.isAbsolute()) {
            return file;
        } else {
            return new File(getHiveMQHomeFolder(), fileLocation);
        }
    }

    private File setUpHiveMQFolder(
            final String propertyName,
            final String variableName,
            final String defaultFolder,
            final boolean createFolderIfNotExists) {
        final String folderName = getSystemPropertyOrEnvironmentVariable(propertyName, variableName);
        final File folder;
        if (folderName != null) {
            folder = findAbsoluteAndRelative(folderName);
        } else {
            folder = findAbsoluteAndRelative(defaultFolder);
        }

        if (createFolderIfNotExists && !folder.exists()) {
            final boolean mkdirsResult = folder.mkdirs();
            if (!mkdirsResult) {
                log.warn("Not able to create folder {}, HiveMQ will behave unexpectedly!", folder);
            }
        }

        return folder;
    }

    /**
     * Returns the value of the system property or environment variable prioritising the system property.
     *
     * @param propertyName the name of the system property
     * @param variableName the name of the environment variable
     * @return value of the system property, if not present value of the environment variable, if not present null
     */
    private String getSystemPropertyOrEnvironmentVariable(
            final String propertyName,
            final String variableName) {
        final String systemProperty = System.getProperty(propertyName);
        if (systemProperty != null) {
            return systemProperty;
        }

        final String environmentVariable = System.getenv().get(variableName);
        return environmentVariable;
    }

    private void setHomeFolder() {
//        final String home =
//                getSystemPropertyOrEnvironmentVariable(SystemProperties.HIVEMQ_HOME, EnvironmentVariables.HIVEMQ_HOME);
//
//        if (home != null) {
//            homeFolder = findAbsoluteAndRelative(home);
//            log.info("HiveMQ home directory: {}", homeFolder.getAbsolutePath());
//
//            //setting system property to support the deprecated PathUtils in the SPI
//            System.setProperty(SystemProperties.HIVEMQ_HOME, homeFolder.getAbsolutePath());
//
//        } else if (usePathOfRunningJar) {
//            usePathOfRunningJarAsHomeFolder();
//        } else {
//            useTemporaryHomeFolder();
//        }
    }

//
//    private void usePathOfRunningJarAsHomeFolder() {
//        final File pathOfRunningJar = getPathOfRunningJar();
//        if (!embedded) {
//            log.warn("No {} property or {} environment variable was set. Using {}",
//                    SystemProperties.HIVEMQ_HOME,
//                    EnvironmentVariables.HIVEMQ_HOME,
//                    pathOfRunningJar.getAbsolutePath());
//        }
//        homeFolder = pathOfRunningJar;
//    }

//    private File getPathOfRunningJar() {
//        final String decode =
//                URLDecoder.decode(HiveMQServer.class.getProtectionDomain().getCodeSource().getLocation().getPath(),
//                        StandardCharsets.UTF_8);
//        final String path = decode.substring(0, decode.lastIndexOf('/') + 1);
//        return new File(path);
//    }

    @Override
    public int getProcessorCount() {
        return this.processorCount;
    }

    public boolean isEmbedded() {
        return embedded;
    }
}
