/*
 * Copyright 2019-present HiveMQ GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.mortals.mqtt.broker.logging;

import com.mortals.mqtt.broker.bootstrap.util.NettyUtil;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import io.netty.channel.Channel;
import org.slf4j.Logger;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;

public class EventLog {

    private static final Logger log = MqttLogger.eventLog;

    private static final DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
    public static final ZoneId ZONE = ZoneId.of("UTC");

    /**
     * 出站消息取消日志记录
     * Log that a outgoing publish message was dropped.
     *
     * @param clientId of the subscriber that would have received the message
     * @param topic    of the publish message
     * @param qos      of the publish message
     * @param reason   why the message was dropped
     */
    public  void messageDropped(final String clientId, final String topic, final int qos, final String reason) {
        log.debug("Outgoing publish message was dropped. Receiving client: {}, topic: {}, qos: {}, reason: {}.",
                valueOrUnknown(clientId), valueOrUnknown(topic), qos, reason);
    }


    /**
     * Log that a outgoing MQTT message for a client was dropped.
     *
     * @param client      identifier of the client that would have received the message
     * @param messageType MQTT message type
     * @param reason      why the message was dropped
     */
    public  void mqttMessageDropped(final String client, final String messageType, final String reason) {
        log.debug("Outgoing MQTT packet was dropped. Receiving client: {}, messageType: {}, reason: {}.",
                valueOrUnknown(client), valueOrUnknown(messageType), reason);
    }

    /**
     * 客户端连接日志
     * Log that a client has successfully connected to the broker.
     *
     * @param channel of the client connection
     */
    public  void clientConnected(final Channel channel) {

        final String clientId = NettyUtil.getClientId(channel);
        final String ip = NettyUtil.getChannelIP(channel);
        final Boolean cleanStart = NettyUtil.getClearStart(channel);
        final Long sessionExpiry = NettyUtil.getSessionExpiry(channel);

        log.debug("Client ID: {}, IP: {}, Clean Start: {}, Session Expiry: {} connected.", valueOrUnknown(clientId), valueOrUnknown(ip), valueOrUnknown(cleanStart), valueOrUnknown(sessionExpiry));
    }

    /**
     * 客户端主动关闭日志.
     *
     * @param channel      of the client connection
     * @param reasonString reason specified by the client for the DISCONNECT
     */
    public  void clientDisconnected(final Channel channel, final String reasonString) {

        final String clientId = NettyUtil.getClientId(channel);
        final String ip = NettyUtil.getChannelIP(channel);
        channel.attr(NettyUtil.DISCONNECT_EVENT_LOGGED).set(true);
        log.debug("Client ID: {}, IP: {} disconnected ungracefully.", valueOrUnknown(clientId), valueOrUnknown(ip));

    }

    /**
     * broker服务端端口连接
     * Log that the connection to the client was closed by the broker.
     *
     * @param channel of the client connection
     * @param reason  why the connection was closed
     */
    public  void clientWasDisconnected(final Channel channel, final String reason) {
        channel.attr(NettyUtil.DISCONNECT_EVENT_LOGGED).set(true);
        final String clientId = NettyUtil.getClientId(channel);
        final String ip = NettyUtil.getChannelIP(channel);
        log.trace("Client {} was disconnected.", clientId);
        log.debug("Client ID: {}, IP: {} was disconnected. reason: {}.", valueOrUnknown(clientId), valueOrUnknown(ip), reason);
    }

    /**
     * Log that a client session expired and will be deleted.
     *
     * @param expiryTimestamp the {@link Long} timestamp of the client-session-expiration
     * @param clientId        of the expired session
     */
    public  void clientSessionExpired(final Long expiryTimestamp, final String clientId) {

        final LocalDateTime disconnectedSinceDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(expiryTimestamp),
                ZONE);
        log.debug("Client ID: {} session has expired at {}. All persistent data for this client has been removed.",
                valueOrUnknown(clientId), disconnectedSinceDateTime.format(dateTimeFormatter));
    }


    private  String valueOrUnknown(final Object object) {
        return object != null ? object.toString() : "UNKNOWN";
    }
}
