package com.mortals.mqtt.broker.processor.protocol;

import com.alibaba.fastjson.JSON;
import com.mortals.mqtt.broker.common.BrokerConstants;
import com.mortals.mqtt.broker.common.code.DeviceOnlineEnum;
import com.mortals.mqtt.broker.common.helper.DateUtils;
import com.mortals.mqtt.broker.common.model.*;
import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.common.queue.DefaultTbQueueMsg;
import com.mortals.common.queue.TbQueueMsg;
import com.mortals.common.queue.TbQueueMsgHeaders;
import com.mortals.common.queue.TbQueueProducer;
import com.mortals.mqtt.broker.services.auth.AuthService;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import com.mortals.common.cluster.dispatcher.event.Event;
import com.mortals.common.cluster.dispatcher.event.EventCode;
import com.mortals.mqtt.broker.processor.receiver.ReSendMessageService;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.bootstrap.util.MessageUtil;
import com.mortals.mqtt.broker.bootstrap.util.NettyUtil;
import com.mortals.common.queue.kafka.DefaultTbQueueMsgHeaders;
import com.mortals.common.queue.TopicPartitionInfo;
import com.mortals.mqtt.broker.store.SessionState;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.store.SubscriptionStore;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.mqtt.*;
import io.netty.handler.timeout.IdleStateHandler;
import org.slf4j.Logger;

import java.util.*;
import java.util.stream.Collectors;

import static io.netty.channel.ChannelFutureListener.CLOSE_ON_FAILURE;
import static io.netty.channel.ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE;
import static io.netty.handler.codec.mqtt.MqttConnectReturnCode.*;

/**
 * mqtt 客户端连接逻辑处理 强制约束：mqtt在未返回conAck之前，
 * 不接收其它任何mqtt协议报文（mqtt协议可以允许）
 * TODO mqtt5 协议支持
 */
public class ConnectProcessor  extends AbstractMessageProcessor implements RequestProcessor {
    private static final Logger log = MqttLogger.clientTraceLog;
    private Boolean debug = true;
    private AuthService authService;
    private ReSendMessageService reSendMessageService;
    private SubscriptionMatcher subscriptionMatcher;
    private SessionStore sessionStore;
    private TbQueueProducer<TbQueueMsg> producer;
    private SubscriptionStore subscriptionStore;

    public ConnectProcessor(MqttBrokerService mqttBrokerService) {
        super(mqttBrokerService);
        this.authService = mqttBrokerService.getAuthService();
        this.reSendMessageService = mqttBrokerService.getReSendMessageService();
        this.subscriptionMatcher = mqttBrokerService.getSubscriptionMatcher();
        this.sessionStore = mqttBrokerService.getSessionStore();
        this.producer = mqttBrokerService.getProducer();
        this.subscriptionStore= mqttBrokerService.getSubscriptionStore();

    }

    @Override
    public void processRequest(ChannelHandlerContext ctx, MqttMessage mqttMessage) {
        MqttConnectMessage connectMessage = (MqttConnectMessage) mqttMessage;

        MqttConnectReturnCode returnCode = null;
        //获取连接协议基本数据
        String clientIdStr = connectMessage.payload().clientIdentifier();
        if (!clientIdVerify(clientIdStr)) {
            returnCode = CONNECTION_REFUSED_IDENTIFIER_REJECTED;
            LogUtil.warn(log, "[CONNECT] -> {} connect failure,returnCode={}", clientIdStr, returnCode);
            abortConnection(ctx, returnCode);//不符合连接版本，断开连接
            return;
        }

        ClientModel clientModel = buildModel(connectMessage);
        String clientId =clientModel.getClientId();

        ClientSession clientSession = null;
        boolean sessionPresent = false;
        try {
            //校验支持协议的版本
            if (isNotProtocolVersion(connectMessage, MqttVersion.MQTT_3_1) && isNotProtocolVersion(connectMessage, MqttVersion.MQTT_3_1_1)) {
                returnCode = MqttConnectReturnCode.CONNECTION_REFUSED_UNACCEPTABLE_PROTOCOL_VERSION;
                LogUtil.warn(log, "[CONNECT] -> {} connect failure,returnCode={}", clientId, returnCode);
                abortConnection(ctx, returnCode);//不符合连接版本，断开连接
                return;
            }

            if (!authentication(clientModel)) {
                returnCode = CONNECTION_REFUSED_BAD_USER_NAME_OR_PASSWORD;
                LogUtil.warn(log, "[CONNECT] -> {} connect failure,returnCode={}", clientId, returnCode);
                abortConnection(ctx, returnCode);
                return;
            }

            // 设置心跳
            int heartbeatSec = connectMessage.variableHeader().keepAliveTimeSeconds();
            if (heartbeatSec > 0 && !keepAlive(clientId, ctx, heartbeatSec)) {
                LogUtil.warn(log, "[CONNECT] -> set heartbeat failure,clientId:{},heartbeatSec:{}", clientId, heartbeatSec);
                throw new Exception("set heartbeat failure");
            }

            //链接成功，更新当前节点订阅树

            Set<Subscription> subscriptions = subscriptionStore.getSubscriptions(clientId);
            LogUtil.debug(log, "[CONNECT]receive subcriptions,clientId:{},subscriptions:{}", clientId, JSON.toJSONString(subscriptions));
            //更新当前订阅树
            for (Subscription subscription : subscriptions) {
                this.subscriptionMatcher.subscribe(subscription);
            }


            //数据库查询是否存在该clientId的session,如果存在，则标识链接时候设置了持久化session,
            SessionState sessionState = sessionStore.getSession(clientId);
           //默认通知其它节点broker需要清当前client会话信息
            boolean notifyClearOtherSession = true;
            if (sessionState.getState() == SessionState.StateEnum.ONLINE) {
                //如果已经存在则删除临时会话
                ClientSession previousClient = ConnectManager.getInstance().getClient(clientId);
                if (previousClient != null) {
                    //当前节点存在clientId之前的session，关闭之前的session，不通知其它节点
                    previousClient.getCtx().close();
                    ConnectManager.getInstance().removeClient(clientId);
                    notifyClearOtherSession = false;
                }
            }

            if (sessionState.getState() == SessionState.StateEnum.NULL) {
                //不存在session 当前节点创建新的clientSession，不通知其它界定啊
                clientSession = new ClientSession(clientId, false, ctx);
                sessionPresent = false;
                notifyClearOtherSession = false;
            } else {
                //session状态为online或者offline,判断客户端连接上来的session设置状态
                if (clientModel.isCleanSession()) {
                    //临时会话，每次 客户端和服务端连接都是 一个新的session（收不到没连接前的旧消息）
                    LogUtil.warn(log, "[CONNECT] -> createNewClientSession,clientId:{},", clientId);
                    clientSession = createNewClientSession(clientId, ctx);
                    sessionPresent = false;
                    notifyClearOtherSession = false;
                } else {
                    //持久化session 断开连接时，session 里保存着发送一半的数据（未确认的数据）和 需要发往这个session的数据
                    LogUtil.warn(log, "[CONNECT] -> reloadClientSession,clientId:{},", clientId);
                    clientSession = reloadClientSession(ctx, clientId);
                    sessionPresent = true;
                }
            }
            //存储 session 会话
            sessionStore.storeSession(clientId, new SessionState(SessionState.StateEnum.ONLINE));
           //是否通知集群中其它节点清理会话事件，为互踢功能
            if (notifyClearOtherSession) {
                Event event = new Event(EventCode.CLEAR_SESSION.getCode(), clientId, System.currentTimeMillis(), MixAll.getLocalIp());
                //集群通知事件
                clusterEventHandler.sendEvent(event);
            }

            // 4. 处理will 消息
            boolean willFlag = connectMessage.variableHeader().isWillFlag();
            if (willFlag) {
                boolean willRetain = connectMessage.variableHeader().isWillRetain();
                int willQos = connectMessage.variableHeader().willQos();
                String willTopic = connectMessage.payload().willTopic();
                byte[] willPayload = connectMessage.payload().willMessageInBytes();
                storeWillMsg(clientId, willRetain, willQos, willTopic, willPayload);
            }
            returnCode = MqttConnectReturnCode.CONNECTION_ACCEPTED;
            NettyUtil.setClientId(ctx.channel(), clientId);
            NettyUtil.setProductId(ctx.channel(),clientModel.getProduct());

            ConnectManager.getInstance().putClient(clientId, clientSession);

            MqttConnAckMessage ackMessage = MessageUtil.getConnectAckMessage(returnCode, sessionPresent);
            ctx.writeAndFlush(ackMessage);
            //离线消息重发
            reConnect2SendMessage(clientId);

            //通过redis更新设备的连接状态
            updateDeviceOnline(clientId, DeviceOnlineEnum.ONLINE.getValue());

            LogUtil.info(log, "[CONNECT] -> {} connect to this mqtt server", clientId);
            //通过kafka发送链接消息
            TopicPartitionInfo info = TopicPartitionInfo.builder().topic(BrokerConstants.MESSAGE_SEND_TOPIC+clientModel.getProduct()).build();
            TbQueueMsgHeaders headers = new DefaultTbQueueMsgHeaders();
            headers.put(MessageHeader.CLIENTID, clientId.getBytes());

            headers.put(MessageHeader.MESSAGEPROTOCOL, new byte[]{MsgTypeEnum.MSG_MQTT.getValue()});
            headers.put(MessageHeader.MESSAGETYPE, new byte[]{(byte) Message.Type.CONNECT.value()});
            headers.put(MessageHeader.TIMESTAMP, String.valueOf(System.currentTimeMillis()).getBytes());
            TbQueueMsg queueMsg = new DefaultTbQueueMsg(UUID.randomUUID(),JSON.toJSONString(clientModel.getExtParamsMap()).getBytes("utf-8"), headers);
            producer.send(info, queueMsg, null);

            LogUtil.debug(log, "[CONNECT] -> currentNode Subs={}", JSON.toJSONString(subscriptionMatcher.getAllNode()));
        } catch (Exception ex) {
            log.error("error",ex);
            LogUtil.error(log, "[CONNECT] -> Service Unavailable: cause={}", ex);
            returnCode = MqttConnectReturnCode.CONNECTION_REFUSED_SERVER_UNAVAILABLE;
            MqttConnAckMessage ackMessage = MessageUtil.getConnectAckMessage(returnCode, sessionPresent);
            ctx.writeAndFlush(ackMessage);
            ctx.close();
        }
    }

    private ClientModel buildModel(MqttConnectMessage connectMessage){
        ClientModel model = new ClientModel();
        String clientIdStr = connectMessage.payload().clientIdentifier();
        String clientId = clientIdStr.substring(0, clientIdStr.indexOf("|"));
        String product = clientId.split("\\.")[0];
        String extParams = clientIdStr.substring(clientIdStr.indexOf("|") + 1, clientIdStr.lastIndexOf("|"));
        Map<String, String> extParamsMap = Arrays.stream(extParams.split(","))
                .map(item -> item.split("="))
                .collect(Collectors.toMap(x -> x[0], y -> y[1]));

        boolean cleanSession = connectMessage.variableHeader().isCleanSession();
        String userName = connectMessage.payload().userName();
        byte[] password = connectMessage.payload().passwordInBytes();

        model.setClientId(clientId);
        model.setProduct(product);
        model.setExtParamsMap(extParamsMap);
        model.setCleanSession(cleanSession);
        model.setUserName(userName);
        model.setPassword(password);
        return model;

    }

    private boolean keepAlive(String clientId, ChannelHandlerContext ctx, int heatbeatSec) {
        if (this.authService.verifyHeartbeatTime(clientId, heatbeatSec)) {
            int keepAlive = (int) (heatbeatSec * 1.5f);
            if (ctx.pipeline().names().contains("idleStateHandler")) {
                ctx.pipeline().remove("idleStateHandler");
            }
            ctx.pipeline().addFirst("idleStateHandler", new IdleStateHandler(keepAlive, 0, 0));
            return true;
        }
        return false;
    }

    private void storeWillMsg(String clientId, boolean willRetain, int willQos, String willTopic, byte[] willPayload) {
        Map<String, Object> headers = new HashMap<>();
        headers.put(MessageHeader.RETAIN, willRetain);
        headers.put(MessageHeader.QOS, willQos);
        headers.put(MessageHeader.TOPIC, willTopic);
        headers.put(MessageHeader.WILL, true);
        Message message = new Message(Message.Type.WILL, headers, willPayload);
        message.setClientId(clientId);
        messageStore.storeWillMessage(clientId, message);
        LogUtil.info(log, "[WillMessageStore] : {} store will message:{}", clientId, message);
    }

    private ClientSession createNewClientSession(String clientId, ChannelHandlerContext ctx) {
        ClientSession clientSession = new ClientSession(clientId, true);
        clientSession.setCtx(ctx);
        //清除当前客户端所有的离线消息
        this.sessionStore.clearSession(clientId, true);
        return clientSession;
    }


    private ClientSession reloadClientSession(ChannelHandlerContext ctx, String clientId) {
        ClientSession clientSession = new ClientSession(clientId, false);
        clientSession.setCtx(ctx);
//        Set<Subscription> subscriptions = subscriptionStore.getSubscriptions(clientId);
//        LogUtil.debug(log, "[CONNECT]receive subcriptions,clientId:{},subscriptions:{}", clientId, JSON.toJSONString(subscriptions));
//        //更新当前订阅树
//        for (Subscription subscription : subscriptions) {
//            this.subscriptionMatcher.subscribe(subscription);
//        }
        return clientSession;
    }

    private void reConnect2SendMessage(String clientId) {
        this.reSendMessageService.put(clientId);
        this.reSendMessageService.wakeUp();
    }

    private boolean authentication(ClientModel clientModel) {
        if (debug) {
            return true;
        }
        return this.authService.authentication(clientModel);
    }

    private boolean onBlackList(String remoteAddr, String clientId) {
        if (debug) {
            return false;
        }
        return this.authService.onBlacklist(remoteAddr, clientId);
    }

    private boolean clientIdVerify(String clientId) {
        if (debug) {
            return true;
        }
        return this.authService.clientIdVerify(clientId);
    }


    private boolean clientVerifyActive(String clientId) {
        if (debug) {
            return true;
        }
        return this.authService.clientVerifyActive(clientId);
    }

    private boolean updateDeviceOnline(String clientId, Integer online) {
        if (debug) {
            return true;
        }
        return this.authService.updateDeviceOnline(clientId, online);
    }


    private boolean isNotProtocolVersion(MqttConnectMessage msg, MqttVersion version) {
        return msg.variableHeader().version() != version.protocolLevel();
    }

    private void abortConnection(ChannelHandlerContext ctx, MqttConnectReturnCode returnCode) {
        MqttConnAckMessage badProto = MqttMessageBuilders.connAck()
                .returnCode(returnCode)
                .sessionPresent(false).build();
        ctx.channel().writeAndFlush(badProto).addListener(FIRE_EXCEPTION_ON_FAILURE);
        ctx.channel().close().addListener(CLOSE_ON_FAILURE);
    }


}
