package com.mortals.mqtt.broker.processor.protocol;

import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.model.Subscription;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.bootstrap.util.NettyUtil;
import com.mortals.mqtt.broker.store.MessageStore;
import com.mortals.mqtt.broker.store.SessionState;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.mqtt.MqttMessage;
import org.slf4j.Logger;

import java.util.Set;

/**
 * 客户端主动发起断开连接：正常断连
 * TODO mqtt5实现
 */
public class DisconnectProcessor implements RequestProcessor {

    private static final Logger log = MqttLogger.clientTraceLog;
    private SessionStore sessionStore;
    private MessageStore messageStore;
    private SubscriptionMatcher subscriptionMatcher;

    public DisconnectProcessor(MqttBrokerService mqttBrokerService) {
        this.sessionStore = mqttBrokerService.getSessionStore();
        this.messageStore = mqttBrokerService.getMessageStore();
        this.subscriptionMatcher = mqttBrokerService.getSubscriptionMatcher();
    }

    @Override
    public void processRequest(ChannelHandlerContext ctx, MqttMessage mqttMessage) {
        String clientId = NettyUtil.getClientId(ctx.channel());
        if (!ConnectManager.getInstance().containClient(clientId)) {
            LogUtil.info(log, "[DISCONNECT] -> {} hasn't connect before", clientId);
        }
        ClientSession clientSession = ConnectManager.getInstance().getClient(clientId);

        // 1. 清理会话 或 重新设置该客户端会话状态
        clearSession(clientSession);

        // 3. 清理will消息
        clearWillMessage(clientSession.getClientId());

        // 4. 移除本节点上的连接
        ConnectManager.getInstance().removeClient(clientId);

        LogUtil.info(log, "[DISCONNECT] -> {}", clientId);

        ctx.close();
    }

    /**
     * 清楚会话，取消订阅
     * @param clientSession
     */
    private void clearSession(ClientSession clientSession) {
        if (clientSession.isCleanStart()) {
            Set<Subscription> subscriptions = sessionStore.getSubscriptions(clientSession.getClientId());
            for (Subscription subscription : subscriptions) {
                this.subscriptionMatcher.unSubscribe(subscription.getTopic(), clientSession.getClientId());
            }

            //更新redis topic信息
            sessionStore.clearSession(clientSession.getClientId(), false);
        } else {
            //持久性会话
            SessionState sessionState = new SessionState(SessionState.StateEnum.OFFLINE, System.currentTimeMillis());
            this.sessionStore.storeSession(clientSession.getClientId(), sessionState);
        }
    }

    private void clearWillMessage(String clientId) {
        messageStore.clearWillMessage(clientId);
    }

}
