package com.mortals.mqtt.broker.processor.protocol;

import com.mortals.mqtt.broker.common.BrokerConstants;
import com.mortals.mqtt.broker.dropping.MessageDroppedService;
import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.mqtt.broker.services.auth.AuthService;
import com.mortals.mqtt.broker.common.helper.DateUtils;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.model.Message;
import com.mortals.mqtt.broker.common.model.MessageHeader;
import com.mortals.mqtt.broker.common.model.MsgTypeEnum;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.bootstrap.util.MessageUtil;
import com.mortals.mqtt.broker.bootstrap.util.NettyUtil;
import com.mortals.mqtt.broker.bootstrap.util.TokenBucketNum;
import com.mortals.common.queue.DefaultTbQueueMsg;
import com.mortals.common.queue.TbQueueMsg;
import com.mortals.common.queue.TbQueueMsgHeaders;
import com.mortals.common.queue.TbQueueProducer;
import com.mortals.common.queue.kafka.DefaultTbQueueMsgHeaders;
import com.mortals.common.queue.TopicPartitionInfo;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.mqtt.MqttMessage;
import io.netty.handler.codec.mqtt.MqttPubAckMessage;
import io.netty.handler.codec.mqtt.MqttPublishMessage;
import io.netty.handler.codec.mqtt.MqttQoS;
import io.netty.util.ReferenceCountUtil;
import org.slf4j.Logger;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * 客户端publish消息到mqtt broker
 *
 */
public class PublishProcessor extends AbstractMessageProcessor implements RequestProcessor {
    private Logger log = MqttLogger.messageTraceLog;

    private AuthService authService;

    private Map<String, TokenBucketNum> map;

    private TbQueueProducer producer;

    private MessageDroppedService messageDroppedService;

    public PublishProcessor(MqttBrokerService controller) {
        super(controller);
        this.map = controller.getTokenBucketMap();
        this.authService = controller.getAuthService();
        this.producer = controller.getProducer();
        this.messageDroppedService = controller.getMessageDroppedService();
    }

    @Override
    public void processRequest(ChannelHandlerContext ctx, MqttMessage mqttMessage) {
        try {

            MqttPublishMessage publishMessage = (MqttPublishMessage) mqttMessage;
            MqttQoS qos = publishMessage.fixedHeader().qosLevel();
            Message innerMsg = new Message();
            String clientId = NettyUtil.getClientId(ctx.channel());
            String productId = NettyUtil.getProductId(ctx.channel());
            ClientSession clientSession = ConnectManager.getInstance().getClient(clientId);
            String topic = publishMessage.variableHeader().topicName();
            if (!this.authService.publishVerify(clientId, topic)) {
                LogUtil.warn(log, "[PubMessage] permission is not allowed");
                clientSession.getCtx().close();
                return;
            }

            //针对clientId进行流控
            TokenBucketNum tokenBucketNum = map.get(clientId);
            if (null != tokenBucketNum) {
                boolean bool = tokenBucketNum.tryAcquire();
                if (!bool) {
                    //丢弃消息，流控
                    messageDroppedService.publishExceedByBucket(clientId, topic, qos.value(), tokenBucketNum);
                    return;
                }
            }

            LogUtil.info(log, "[PubMessage] -> receive mqtt pub message topic :{}", topic);
            innerMsg.setPayload(MessageUtil.readBytesFromByteBuf(((MqttPublishMessage) mqttMessage).payload()));
            innerMsg.setClientId(clientId);
            innerMsg.setType(Message.Type.valueOf(mqttMessage.fixedHeader().messageType().value()));
            Map<String, Object> headers = new HashMap<>();
            headers.put(MessageHeader.TOPIC, publishMessage.variableHeader().topicName());
            headers.put(MessageHeader.QOS, publishMessage.fixedHeader().qosLevel().value());
            headers.put(MessageHeader.RETAIN, publishMessage.fixedHeader().isRetain());
            headers.put(MessageHeader.DUP, publishMessage.fixedHeader().isDup());
            innerMsg.setHeaders(headers);
            innerMsg.setMsgId(publishMessage.variableHeader().packetId());
            innerMsg.setStoreTime(DateUtils.getCurrDate().getTime());
            switch (qos) {
                case AT_MOST_ONCE:
                    processMessage(innerMsg);
                    break;
                case AT_LEAST_ONCE:
                    processQos1(ctx, innerMsg);
                    break;
                case EXACTLY_ONCE:
                    processQos2(ctx, innerMsg);
                    break;
                default:
                    LogUtil.warn(log, "[PubMessage] -> Wrong mqtt message,clientId={}", clientId);
            }


            //通过消息组件发送消息
            TopicPartitionInfo info = TopicPartitionInfo.builder().topic(BrokerConstants.MESSAGE_SEND_TOPIC + productId).build();
            TbQueueMsgHeaders header = new DefaultTbQueueMsgHeaders();
            header.put(MessageHeader.CLIENTID, clientId.getBytes());
            header.put(MessageHeader.MESSAGEPROTOCOL, new byte[]{MsgTypeEnum.MSG_MQTT.getValue()});
            header.put(MessageHeader.MESSAGETYPE, new byte[]{(byte) Message.Type.PUBLISH.value()});
            header.put(MessageHeader.TOPIC, topic.getBytes());
            header.put(MessageHeader.QOS, new byte[]{(byte) qos.value()});
            headers.put(MessageHeader.TIMESTAMP, DateUtils.getCurrStrDateTime().getBytes());


            TbQueueMsg queueMsg = new DefaultTbQueueMsg(UUID.randomUUID(), innerMsg.getPayload(), header);
            producer.send(info, queueMsg, null);
        } catch (Throwable tr) {
            LogUtil.warn(log, "[PubMessage] -> Solve mqtt pub message exception:{}", tr.getMessage());
        } finally {
            ReferenceCountUtil.release(mqttMessage.payload());
        }
    }

    private void processQos2(ChannelHandlerContext ctx, Message innerMsg) {
        int originMessageId = innerMsg.getMsgId();
        LogUtil.debug(log, "[PubMessage] -> Process qos2 message,clientId={}", innerMsg.getClientId());
        boolean flag = cacheInflowMsg(innerMsg.getClientId(), innerMsg);
        if (!flag) {
            LogUtil.warn(log, "[PubMessage] -> cache qos2 pub message failure,clientId={}", innerMsg.getClientId());
        }
        MqttMessage pubRecMessage = MessageUtil.getPubRecMessage(originMessageId);
        ctx.writeAndFlush(pubRecMessage);
    }

    private void processQos1(ChannelHandlerContext ctx, Message innerMsg) {
        int originMessageId = innerMsg.getMsgId();
        processMessage(innerMsg);
        LogUtil.info(log, "[PubMessage] -> Process qos1 message,clientId={}", innerMsg.getClientId());
        MqttPubAckMessage pubAckMessage = MessageUtil.getPubAckMessage(originMessageId);
        ctx.writeAndFlush(pubAckMessage);
    }

}
