package com.mortals.mqtt.broker.processor.protocol;

import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.mqtt.broker.services.auth.AuthService;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.model.Message;
import com.mortals.mqtt.broker.common.model.MessageHeader;
import com.mortals.mqtt.broker.common.model.Subscription;
import com.mortals.mqtt.broker.common.model.Topic;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import com.mortals.mqtt.broker.bootstrap.session.ClientSession;
import com.mortals.mqtt.broker.bootstrap.session.ConnectManager;
import com.mortals.mqtt.broker.bootstrap.util.MessageUtil;
import com.mortals.mqtt.broker.bootstrap.util.NettyUtil;
import com.mortals.mqtt.broker.store.MessageStore;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.store.SubscriptionStore;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.mqtt.MqttMessage;
import io.netty.handler.codec.mqtt.MqttPublishMessage;
import io.netty.handler.codec.mqtt.MqttSubscribeMessage;
import io.netty.handler.codec.mqtt.MqttTopicSubscription;
import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * 订阅报文逻辑处理
 * TODO mqtt5协议支持
 */
public class SubscribeProcessor implements RequestProcessor {

    private static final Logger log = MqttLogger.messageTraceLog;

    private SubscriptionMatcher subscriptionMatcher;
    private AuthService authService;
    private MessageStore messageStore;
    private SessionStore sessionStore;
    private SubscriptionStore subscriptionStore;

    public SubscribeProcessor(MqttBrokerService controller) {
        this.subscriptionMatcher = controller.getSubscriptionMatcher();
        this.authService = controller.getAuthService();
        this.sessionStore = controller.getSessionStore();
        this.messageStore = controller.getMessageStore();
        this.subscriptionStore = controller.getSubscriptionStore();
    }

    @Override
    public void processRequest(ChannelHandlerContext ctx, MqttMessage mqttMessage) {
        MqttSubscribeMessage subscribeMessage = (MqttSubscribeMessage) mqttMessage;
        String clientId = NettyUtil.getClientId(ctx.channel());
        int messageId = subscribeMessage.variableHeader().messageId();
        ClientSession clientSession = ConnectManager.getInstance().getClient(clientId);
        List<Topic> validTopicList = validTopics(clientSession, subscribeMessage.payload().topicSubscriptions());
        if (validTopicList == null || validTopicList.size() == 0) {
            LogUtil.warn(log, "[Subscribe] -> Valid all subscribe topic failure,clientId:{}", clientId);
            return;
        }
        List<Integer> ackQos = getTopicQos(validTopicList);
        MqttMessage subAckMessage = MessageUtil.getSubAckMessage(messageId, ackQos);
        ctx.writeAndFlush(subAckMessage);
        //获取留存消息，并更新订阅消息
        List<Message> retainMessages = subscribe(clientSession, validTopicList);
        //分发留存消息
        dispatcherRetainMessage(clientSession, retainMessages);

        //集群分发订阅事件，各个节点更新订阅树
    }

    private List<Integer> getTopicQos(List<Topic> topics) {
        return topics.stream().map(item -> item.getQos()).collect(Collectors.toList());

    }

    private List<Message> subscribe(ClientSession clientSession, List<Topic> validTopicList) {
        Collection<Message> retainMessages = null;
        List<Message> needDispatcher = new ArrayList<>();
        for (Topic topic : validTopicList) {
            Subscription subscription = new Subscription(clientSession.getClientId(), topic.getTopicName(), topic.getQos());
            //将订阅者添加到订阅树中
            boolean subRs = this.subscriptionMatcher.subscribe(subscription);

            //留存消息处理
            if (subRs) {
                if (retainMessages == null) {
                    // TODO 这里需要优化，不能一次获取所有retain消息，
                    //  retain消息太多可能导致broker crash或者hang住
                    retainMessages = messageStore.getAllRetainMsg();
                }
                if (!MixAll.isEmpty(retainMessages)) {
                    for (Message retainMsg : retainMessages) {
                        String pubTopic = (String) retainMsg.getHeader(MessageHeader.TOPIC);
                        if (subscriptionMatcher.isMatch(pubTopic, subscription.getTopic())) {
                            int minQos = MessageUtil.getMinQos((int) retainMsg.getHeader(MessageHeader.QOS), topic.getQos());
                            retainMsg.putHeader(MessageHeader.QOS, minQos);
                            needDispatcher.add(retainMsg);
                        }
                    }
                }
            }
            this.sessionStore.storeSubscription(clientSession.getClientId(), subscription);
            this.subscriptionStore.storeSubscription(clientSession.getClientId(), subscription);
        }
        return needDispatcher;
    }

    /**
     * 返回校验合法的topic
     */
    private List<Topic> validTopics(ClientSession clientSession, List<MqttTopicSubscription> topics) {
        List<Topic> topicList = new ArrayList<>();
        for (MqttTopicSubscription subscription : topics) {
            if (!authService.subscribeVerify(clientSession.getClientId(), subscription.topicName())) {
                LogUtil.warn(log, "[SubPermission] this clientId:{} have no permission to subscribe this topic:{}", clientSession.getClientId(), subscription.topicName());
                clientSession.getCtx().close();
                return null;
            }
            Topic topic = new Topic(subscription.topicName(), subscription.qualityOfService().value());
            topicList.add(topic);
        }
        return topicList;
    }

    /**
     * 分发retain消息:
     * TODO 待优化，retain消息逻辑需要优化：1.性能优化；2.逻辑放到MessageDispatcher统一处理
     */
    private void dispatcherRetainMessage(ClientSession clientSession, List<Message> messages) {
        for (Message message : messages) {
            message.putHeader(MessageHeader.RETAIN, true);
            int qos = (int) message.getHeader(MessageHeader.QOS);
            if (qos > 0) {
                sessionStore.cacheInflowMsg(clientSession.getClientId(), message);
            }
            message.setMsgId(clientSession.generateMessageId());
            MqttPublishMessage publishMessage = MessageUtil.getPubMessage(message, false);
            clientSession.getCtx().writeAndFlush(publishMessage);
        }
    }

}
