package com.mortals.mqtt.broker.services;

import com.codahale.metrics.ConsoleReporter;
import com.codahale.metrics.MetricRegistry;
import com.mortals.common.scheduler.SchedulerComponent;
import com.mortals.mqtt.broker.dropping.MessageDroppedService;
import com.mortals.mqtt.broker.dropping.MessageDroppedServiceImpl;
import com.mortals.mqtt.broker.logging.EventLog;
import com.mortals.mqtt.broker.metrics.common.MetricsHolder;
import com.mortals.mqtt.broker.services.device.DeviceServiceImpl;
import com.mortals.mqtt.broker.services.auth.AuthService;
import com.mortals.mqtt.broker.services.device.DeviceService;
import com.mortals.mqtt.broker.bootstrap.ClientLifeCycleHookService;
import com.mortals.mqtt.broker.common.config.BrokerConfig;
import com.mortals.mqtt.broker.common.config.NettyConfig;
import com.mortals.mqtt.broker.common.helper.MixAll;
import com.mortals.mqtt.broker.common.helper.RejectHandler;
import com.mortals.mqtt.broker.common.helper.ThreadFactoryImpl;
import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.model.Subscription;
import com.mortals.mqtt.broker.processor.RequestProcessor;
import com.mortals.common.cluster.dispatcher.ClusterEventHandler;
import com.mortals.common.cluster.dispatcher.DefaultDispatcherInnerMessage;
import com.mortals.common.cluster.dispatcher.EventConsumeHandler;
import com.mortals.common.cluster.dispatcher.InnerMessageDispatcher;
import com.mortals.mqtt.broker.processor.protocol.*;
import com.mortals.mqtt.broker.processor.receiver.ReSendMessageService;
import com.mortals.mqtt.broker.bootstrap.netty.ChannelEventListener;
import com.mortals.mqtt.broker.bootstrap.netty.NettyNettyServer;
import com.mortals.mqtt.broker.bootstrap.util.TokenBucketNum;
import com.mortals.common.scheduler.DefaultSchedulerComponent;
import com.mortals.common.queue.ConsumerService;
import com.mortals.common.queue.MessageQueueFactory;
import com.mortals.common.queue.TbQueueMsg;
import com.mortals.common.queue.TbQueueProducer;
import com.mortals.mqtt.broker.store.MessageStore;
import com.mortals.mqtt.broker.store.SessionStore;
import com.mortals.mqtt.broker.store.SubscriptionStore;
import com.mortals.mqtt.broker.store.memory.InflowMessageHandler;
import com.mortals.mqtt.broker.store.memory.OutflowMessageHandler;
import com.mortals.mqtt.broker.store.memory.OutflowSecMessageHandler;
import com.mortals.mqtt.broker.store.redis.RedisSubscriptionStore;
import com.mortals.mqtt.broker.subscribe.DefaultSubscriptionTreeMatcher;
import com.mortals.mqtt.broker.subscribe.SubscriptionMatcher;
import io.netty.handler.codec.mqtt.MqttMessageType;
import lombok.Data;
import org.slf4j.Logger;

import java.util.Map;
import java.util.Set;
import java.util.concurrent.*;
import java.util.stream.Collectors;

/**
 * 具体控制类：负责加载配置文件，初始化环境，启动服务，kafka初始化等
 */
@Data
public class MqttBrokerService {

    private static final Logger log = MqttLogger.brokerlog;

    private BrokerConfig brokerConfig;
    private NettyConfig nettyConfig;


    private volatile Map<String, TokenBucketNum> tokenBucketMap = new ConcurrentHashMap<>();

    //连接执行线程服务
    private ExecutorService connectExecutor;
    //发布执行线程服务
    private ExecutorService pubExecutor;
    //订阅执行线程服务
    private ExecutorService subExecutor;
    //心跳执行线程服务
    private ExecutorService pingExecutor;

    //连接队列
    private LinkedBlockingQueue<Runnable> connectQueue;
    //发布队列
    private LinkedBlockingQueue<Runnable> pubQueue;
    //订阅队列
    private LinkedBlockingQueue<Runnable> subQueue;
    //心跳 队列
    private LinkedBlockingQueue<Runnable> pingQueue;
    //通道事件监听
    private ChannelEventListener channelEventListener;
    //netty远程服务器
    private NettyNettyServer nettyServer;
    //内部消息分发器
    private InnerMessageDispatcher innerMessageDispatcher;
    //订阅匹配服务
    private SubscriptionMatcher subscriptionMatcher;
    //鉴权认证
    private AuthService authService;

    private DeviceService deviceService;
    //消息重发服务
    private ReSendMessageService reSendMessageService;
    //session存取
    private SessionStore sessionStore;
    //消息存取
    private MessageStore messageStore;

    private SubscriptionStore subscriptionStore;
    //集群事件handler
    private ClusterEventHandler clusterEventHandler;
    //集群消费事件handler
    private EventConsumeHandler eventConsumeHandler;
    private String currentIp;
    // high performance message handle
    //消息出入栈以内存存取方式
    private InflowMessageHandler inflowMessageHandler;
    private OutflowMessageHandler outflowMessageHandler;
    private OutflowSecMessageHandler outflowSecMessageHandler;

    private MessageQueueFactory messageQueueFactory;

    //消息生产服务
    private TbQueueProducer<TbQueueMsg> producer;
    //消息消费类
    private ConsumerService consumerService;
    //周期任务
    private DefaultSchedulerComponent schedulerComponent;
    //消息丢弃服务
    private MessageDroppedService messageDroppedService;
    //度量类
    private MetricsHolder metricsHolder;


    /**
     * 初始化数据
     *
     * @param brokerConfig
     * @param nettyConfig
     */
    public MqttBrokerService(BrokerConfig brokerConfig, NettyConfig nettyConfig) {

        this.brokerConfig = brokerConfig;
        this.nettyConfig = nettyConfig;
        //初始化队列大小
        this.connectQueue = new LinkedBlockingQueue<>(100000);
        this.pubQueue = new LinkedBlockingQueue<>(100000);
        this.subQueue = new LinkedBlockingQueue<>(100000);
        this.pingQueue = new LinkedBlockingQueue<>(10000);
        //获取ip地址
        this.currentIp = MixAll.getLocalIp();

        this.metricsHolder = new MetricsHolder(brokerConfig.getMetricRegistry());
        //消息丢弃服务
        this.messageDroppedService = new MessageDroppedServiceImpl(this.metricsHolder, new EventLog());

        // 会话状态，消息存储加载，可自己实现相关的类
        //反射加载方式 配置的类路径地址
        this.sessionStore = MixAll.pluginInit(brokerConfig.getSessionStoreClass());
        this.messageStore = MixAll.pluginInit(brokerConfig.getMessageStoreClass());

        this.subscriptionStore = new RedisSubscriptionStore();
        this.deviceService = new DeviceServiceImpl();

        // 设备连接，发布，订阅消息权限控制
        this.authService = MixAll.pluginInit(brokerConfig.getAuthValidClass());
        // 集群事件转发加载
        this.clusterEventHandler = MixAll.pluginInit(brokerConfig.getClusterEventHandlerClass());

        // high performance message handler
        this.inflowMessageHandler = new InflowMessageHandler();
        this.outflowMessageHandler = new OutflowMessageHandler();
        this.outflowSecMessageHandler = new OutflowSecMessageHandler();
        //订阅匹配
        this.subscriptionMatcher = new DefaultSubscriptionTreeMatcher();
        this.innerMessageDispatcher = new DefaultDispatcherInnerMessage(this);
        this.eventConsumeHandler = new EventConsumeHandler(this);

        this.channelEventListener = new ClientLifeCycleHookService(messageStore, sessionStore, innerMessageDispatcher);
        this.nettyServer = new NettyNettyServer(brokerConfig, nettyConfig, channelEventListener,metricsHolder,messageDroppedService);
        this.reSendMessageService = new ReSendMessageService(this);

        this.messageQueueFactory = MixAll.pluginInit(brokerConfig.getMsgQueueClass());

        this.producer = messageQueueFactory.createMsgProducer(brokerConfig);

        this.schedulerComponent = new DefaultSchedulerComponent();

        //初始化线程池
        int coreThreadNum = Runtime.getRuntime().availableProcessors();
        this.connectExecutor = new ThreadPoolExecutor(coreThreadNum * 2,
                coreThreadNum * 2,
                60000,
                TimeUnit.MILLISECONDS,
                connectQueue,
                new ThreadFactoryImpl("ConnectThread"),
                new RejectHandler("connect", 100000));
        this.pubExecutor = new ThreadPoolExecutor(coreThreadNum * 2,
                coreThreadNum * 2,
                60000,
                TimeUnit.MILLISECONDS,
                pubQueue,
                new ThreadFactoryImpl("PubThread"),
                new RejectHandler("pub", 100000));
        this.subExecutor = new ThreadPoolExecutor(coreThreadNum * 2,
                coreThreadNum * 2,
                60000,
                TimeUnit.MILLISECONDS,
                subQueue,
                new ThreadFactoryImpl("SubThread"),
                new RejectHandler("sub", 100000));
        this.pingExecutor = new ThreadPoolExecutor(coreThreadNum,
                coreThreadNum,
                60000,
                TimeUnit.MILLISECONDS,
                pingQueue,
                new ThreadFactoryImpl("PingThread"),
                new RejectHandler("heartbeat", 100000));

    }


    public void start() {

        //打印配置参数
        MixAll.printProperties(log, brokerConfig);
        MixAll.printProperties(log, nettyConfig);


        // 1. start store
        this.sessionStore.start(brokerConfig);
        this.messageStore.start(brokerConfig);
        this.subscriptionStore.start(brokerConfig);


        // 2. start cluster
        this.eventConsumeHandler.start();
        this.clusterEventHandler.start(brokerConfig);

        // 3. start message service
        if (this.innerMessageDispatcher != null) {
            this.innerMessageDispatcher.start();
        }
        //消息重发
        if (this.reSendMessageService != null) {
            this.reSendMessageService.start();
        }


        // 4. init and register mqtt protocol processor
        RequestProcessor connectProcessor = new ConnectProcessor(this);
        RequestProcessor disconnectProcessor = new DisconnectProcessor(this);
        RequestProcessor pingProcessor = new PingProcessor();
        RequestProcessor publishProcessor = new PublishProcessor(this);
        RequestProcessor pubRelProcessor = new PubRelProcessor(this);
        RequestProcessor subscribeProcessor = new SubscribeProcessor(this);
        RequestProcessor unSubscribeProcessor = new UnSubscribeProcessor(subscriptionMatcher, sessionStore);
        RequestProcessor pubRecProcessor = new PubRecProcessor(this);
        RequestProcessor pubAckProcessor = new PubAckProcessor(this);
        RequestProcessor pubCompProcessor = new PubCompProcessor(this);

        this.nettyServer.registerProcessor(MqttMessageType.CONNECT, connectProcessor, connectExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.DISCONNECT, disconnectProcessor, connectExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PINGREQ, pingProcessor, pingExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PUBLISH, publishProcessor, pubExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PUBACK, pubAckProcessor, pubExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PUBREL, pubRelProcessor, pubExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.SUBSCRIBE, subscribeProcessor, subExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.UNSUBSCRIBE, unSubscribeProcessor, subExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PUBREC, pubRecProcessor, subExecutor);
        this.nettyServer.registerProcessor(MqttMessageType.PUBCOMP, pubCompProcessor, subExecutor);

        // 5. start auth
        if (this.authService != null) {
            this.authService.start();
        }

        // 6. start remoting
        if (this.nettyServer != null) {
            this.nettyServer.start();
        }
        LogUtil.info(log, "Mqtt Server start success and version = {}", brokerConfig.getVersion());

        //初始化订阅树
        //initSubs();

        //获取所有产品编码，订阅所有产品topic

        //this.consumerService.getMainConsumer().subscribe();
        //初始化消息队列

        this.consumerService = new ConsumerService(brokerConfig.getPollDuration(), this.clusterEventHandler, this.currentIp);

        this.consumerService.init(messageQueueFactory.createMsgConsumer(brokerConfig));

        //初始化令牌桶
        this.deviceService.getAllDeviceModel().parallelStream().peek(item -> {
            TokenBucketNum tokenBucket = TokenBucketNum.newBuilder().avgFlowRate(item.getAvgFlowRate()).maxFlowRate(item.getMaxFlowRate()).build();
            tokenBucketMap.put(item.getClientId(), tokenBucket);
        }).count();


        //控制台报告
        ConsoleReporter reporter = ConsoleReporter.forRegistry(metricsHolder.getMetricRegistry())
                .convertRatesTo(TimeUnit.SECONDS)
                .convertDurationsTo(TimeUnit.MILLISECONDS)
                .build();
        reporter.start(60, TimeUnit.SECONDS);

        Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
            @Override
            public void run() {
                shutdown();
            }
        }));
    }

//    private void initSubs() {
//        Set<String> allDeviceClients = this.deviceService.getAllDeviceClients();
//        //根据设备获取设备所有订阅topic
//        Set<Subscription> allSubs = allDeviceClients.parallelStream()
//                .flatMap(item -> this.subscriptionStore.getSubscriptions(item).stream()).collect(Collectors.toSet());
//        //根据订阅的top,初始化订阅树
//        allSubs.parallelStream().map(item -> subscriptionMatcher.subscribe(item)).count();
//    }

    public void shutdown() {


        if (this.producer != null) {
            this.producer.stop();
        }

        if (this.consumerService != null) {
            this.consumerService.destroy();
        }
        if (this.nettyServer != null) {
            this.nettyServer.shutdown();
        }
        if (this.connectExecutor != null) {
            this.connectExecutor.shutdown();
        }
        if (this.pubExecutor != null) {
            this.pubExecutor.shutdown();
        }
        if (this.subExecutor != null) {
            this.subExecutor.shutdown();
        }
        if (this.pingExecutor != null) {
            this.pingExecutor.shutdown();
        }
        if (this.innerMessageDispatcher != null) {
            this.innerMessageDispatcher.shutdown();
        }
        if (this.reSendMessageService != null) {
            this.reSendMessageService.shutdown();
        }
        if (this.clusterEventHandler != null) {
            this.clusterEventHandler.shutdown();
        }
        if (this.eventConsumeHandler != null) {
            this.eventConsumeHandler.shutdown();
        }
        if (this.sessionStore != null) {
            this.sessionStore.shutdown();
        }
        if (this.messageStore != null) {
            this.messageStore.shutdown();
        }

        if (this.messageStore != null) {
            this.messageStore.shutdown();
        }

        if (this.subscriptionStore != null) {
            this.subscriptionStore.shutdown();
        }

        if (this.authService != null) {
            this.authService.shutdown();
        }

        if (this.schedulerComponent != null) {
            this.schedulerComponent.destroy();
        }

    }
}
