package com.mortals.mqtt.broker.subscribe;

import com.mortals.mqtt.broker.common.log.LogUtil;
import com.mortals.mqtt.broker.common.log.MqttLogger;
import com.mortals.mqtt.broker.common.model.Subscription;
import lombok.Getter;
import org.slf4j.Logger;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class DefaultSubscriptionTreeMatcher implements SubscriptionMatcher {

    private static final Logger log = MqttLogger.messageTraceLog;

    private static final String GROUP_STR = "$share";
    private final Object lock = new Object();
    @Getter
    private TreeNode root = new TreeNode(new Token("root"));
    private Token EMPTY = new Token("");
    private Token SINGLE = new Token("+");
    private Token MULTY = new Token("#");
    private Token GROUP = new Token(GROUP_STR);

    //TODO 共享订阅均衡策略（提取到配置文件） mqtt5.0
    private String sharedSubscriptionStrategy = "random";


    public DefaultSubscriptionTreeMatcher() {
    }


    /**
     * 新增订阅
     * @param subscription
     * @return true：new subscribe,dispatcher retain message
     *         false：no need to dispatcher retain message
     *         是否保存更新retain message
     */
    @Override
    public boolean subscribe(Subscription subscription) {
        try {
            String topic = subscription.getTopic();
            //通过topic和根节点 获取当前节点
            TreeNode currentNode = recursionGetTreeNode(topic, root);
            //获取当前节点的订阅者
            Set<Subscription> subscriptions = currentNode.getSubscribers();
           //判断新增订阅者是否已经存在于当前节点订阅者中
            if (subscriptions.contains(subscription)) {
                for (Subscription sub : subscriptions) {
                    //更新qos质量
                    if (sub.equals(subscription)) {
                        if (sub.getQos() == subscription.getQos()) {
                            return false;
                        } else {
                            sub.setQos(subscription.getQos());
                            return true;
                        }
                    }
                }
            }
            //直接添加
            currentNode.addSubscriber(subscription);
        } catch (Exception ex) {
            LogUtil.warn(log, "[Subscription] -> Subscribe failed,clientId={},topic={},qos={}",
                    subscription.getClientId(), subscription.getTopic(), subscription.getQos());
            return true;
        }
        return true;
    }

    @Override
    public boolean unSubscribe(String topic, String clientId) {
        //退订，删除当前节点下制定的订阅者
        TreeNode currentNode = recursionGetTreeNode(topic, root);
        currentNode.getSubscribers().remove(new Subscription(clientId, topic, 1));
        return true;
    }

    private TreeNode recursionGetTreeNode(String topic, TreeNode node) {
       //分解topic字符串，递归创建树节点。订阅不支持#,+等通配符
        String[] tokens = topic.split("/");
       // System.out.println("tokens:"+JSON.toJSONString(tokens));
        Token token = new Token(tokens[0]);
        //System.out.println(JSON.toJSONString(token));
        TreeNode matchNode = node.getChildNodeByToken(token);
        if (Objects.isNull(matchNode)) {

            synchronized (lock) {
                matchNode = node.getChildNodeByToken(token);
                if (Objects.isNull(matchNode)) {
                    matchNode = new TreeNode(token);
                    //System.out.println("matchNode:"+ JSON.toJSONString(matchNode));
                    node.addChild(matchNode);
                }
            }
        }
        if (tokens.length > 1) {
            String childTopic = topic.substring(topic.indexOf("/") + 1);
            return recursionGetTreeNode(childTopic, matchNode);
        } else {
            return matchNode;
        }
    }

    @Override
    public Set<Subscription> match(String topic, String clientId) {
        Set<Subscription> subscriptions = new HashSet<>();
        recursionMatch(topic, root, false, subscriptions);
        //过滤重复的共享订阅者
        return filterSharedSub(subscriptions, clientId);
    }

    private Set<Subscription> filterSharedSub(Set<Subscription> subscriptions, String clientId) {
        Map<String, Set<Subscription>> groupSharedSubs = new HashMap<>();
        Set<Subscription> filteredSubs = new HashSet<>();
        for (Subscription sub : subscriptions) {
            String topic = sub.getTopic();
            //共享订阅者按topic字串分组
            if (topic.startsWith(GROUP_STR)) {
                //TODO 待定，不同topic字串但匹配同一topic
                Set<Subscription> subs = groupSharedSubs.get(topic);
                if (Objects.isNull(subs)) {
                    subs = new HashSet<>();
                    groupSharedSubs.put(topic, subs);
                }
                subs.add(sub);
            } else {
                filteredSubs.add(sub);
            }
        }

        //按策略从不同订阅组选取订阅者
        Collection<Set<Subscription>> groups = groupSharedSubs.values();
        for (Set<Subscription> group : groups) {
            filteredSubs.add(selectOneByStrategy(group, clientId));
        }
        return filteredSubs;
    }

    /**
     * TODO 按照均衡策略选择一个订阅者;
     * <pre>
     * 策略：
     * random	在所有订阅者中随机选择;
     * hash	按照发布者 ClientID 的哈希值;
     * </pre>
     */
    private Subscription selectOneByStrategy(Set<Subscription> group, String clientId) {
        return group.iterator().next();
    }

    @Override
    public boolean isMatch(String pubTopic, String subTopic) {
        if (pubTopic.equals(subTopic)) {
            return true;
        }
        Pattern pattern = Pattern.compile("(" + GROUP + "/\\S/)(\\S+)");
        Matcher m = pattern.matcher(subTopic);
        if (m.find()) {
            return innerIsMatch(pubTopic, m.group(2));
        }

        return innerIsMatch(pubTopic, subTopic);
    }

    @Override
    public TreeNode getAllNode() {
        return root;
    }

    @Override
    public void initNode() {
        root = new TreeNode(new Token("root"));
    }

    private boolean innerIsMatch(String pubTopic, String subTopic) {
        if (pubTopic.equals(subTopic)) {
            return true;
        }
        String[] pubTokenStr = pubTopic.split("/");
        String[] subTokenStr = subTopic.split("/");
        int pubLen = pubTokenStr.length;
        int subLen = subTokenStr.length;
        if (pubLen != subLen) {
            Token lastSubToken = new Token(subTokenStr[subLen - 1]);
            if (subLen > pubLen || (!lastSubToken.equals(MULTY))) {
                return false;
            }
        }
        for (int i = 0; i < pubLen; i++) {
            Token pubToken = new Token(pubTokenStr[i]);
            Token subToken = new Token(subTokenStr[i]);
            if (subToken.equals(MULTY)) {
                return true;
            } else if ((!pubToken.equals(subToken)) && (!subToken.equals(SINGLE))) {
                return false;
            }
            if (i == subLen - 1) {
                return true;
            }
        }
        return false;
    }

    private void recursionMatch(String topic, TreeNode node, boolean isGroupToken,
                                Set<Subscription> subscriptions) {
        String[] topics = topic.split("/");
        Token token = new Token(topics[0]);
        List<TreeNode> childNodes = node.getChildren();
        if (topics.length > 1) {
            String nextTopic = topic.substring(topic.indexOf("/") + 1);
            for (TreeNode itemNode : childNodes) {

                if (isGroupToken) {
                    recursionMatch(topic, itemNode, false, subscriptions);
                }
                if (itemNode.getToken().equals(token) || itemNode.getToken().equals(SINGLE)) {
                    recursionMatch(nextTopic, itemNode, false, subscriptions);
                }
                if (itemNode.getToken().equals(GROUP)) {
                    recursionMatch(topic, itemNode, true, subscriptions);
                }
                if (itemNode.getToken().equals(MULTY)) {
                    subscriptions.addAll(itemNode.getSubscribers());
                }
            }
        } else {
            for (TreeNode itemNode : childNodes) {
                if (itemNode.getToken().equals(token) || itemNode.getToken().equals(SINGLE)
                        || itemNode.getToken().equals(MULTY)) {
                    subscriptions.addAll(itemNode.getSubscribers());
                }
            }
        }
    }
}
