/*
 * Copyright (c) 2012-2018 The original author or authors
 * ------------------------------------------------------
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Apache License v2.0 which accompanies this distribution.
 *
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * The Apache License v2.0 is available at
 * http://www.opensource.org/licenses/apache2.0.php
 *
 * You may elect to redistribute this code under either of these licenses.
 */

package mortals.mqtt;

import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.MQServer;
import com.mortals.mqtt.broker.common.util.MqttSign;
import org.awaitility.Awaitility;
import org.awaitility.Durations;
import org.eclipse.paho.client.mqttv3.*;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.eclipse.paho.client.mqttv3.persist.MqttDefaultFilePersistence;
import org.junit.jupiter.api.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.TimeUnit;

import static java.nio.charset.StandardCharsets.UTF_8;
import static org.junit.jupiter.api.Assertions.*;

public class ServerIntegrationPahoTest {

    private static final Logger LOG = LoggerFactory.getLogger(ServerIntegrationPahoTest.class);

    IMqttClient m_client;
    IMqttClient m_publisher;
    IMqttClient subscriberA;
    IMqttClient subscriberB;
    MessageCollector m_messagesCollector;

    MqttBrokerService mqttBrokerService;

    public static final String HOST = "tcp://127.0.0.1:1883";
    public static final String HOST1 = "tcp://127.0.0.1:1885";
    public static final String TOPIC = "toclient/124";
    private static final String clientid = "client124";
    private String productKey = "cdbao";
    private String deviceName = "device1";
    private String deviceSecret = "ga7XA6KdlEeiPXQPpRbAjOZXwG8y";


    private String deviceName2 = "device2";

    private String deviceName3 = "device3";

    private String deviceName4 = "device4";


    private MqttSign sign;
    private MqttSign sign2;
    private MqttSign sign3;
    private MqttSign sign4;

    @BeforeAll
    public static void beforeTests() {
        Awaitility.setDefaultTimeout(Durations.ONE_SECOND);
    }

    @BeforeEach
    public void setUp() throws Exception {

        mqttBrokerService = MQServer.startMqttService(new String[]{});

        // 计算MQTT连接参数。
        sign = new MqttSign();
        sign.calculate(productKey, deviceName, deviceSecret);

        sign2 = new MqttSign();
        sign2.calculate(productKey, deviceName2, deviceSecret);

        sign3 = new MqttSign();
        sign3.calculate(productKey, deviceName3, deviceSecret);

        sign4 = new MqttSign();
        sign4.calculate(productKey, deviceName4, deviceSecret);

        m_client = new MqttClient(HOST, sign.getClientid(), new MemoryPersistence());


        m_messagesCollector = new MessageCollector();
        m_client.setCallback(m_messagesCollector);

        m_publisher = new MqttClient(HOST, sign2.getClientid(), new MqttDefaultFilePersistence());
    }

    @AfterEach
    public void tearDown() throws Exception {
        if (m_client != null && m_client.isConnected()) {
            m_client.disconnect();
            m_client.close();
        }

        if (m_publisher != null && m_publisher.isConnected()) {
            m_publisher.disconnect();
            m_publisher.close();
        }


        if (subscriberA != null && subscriberA.isConnected()) {
            subscriberA.disconnect();
            subscriberA.close();
        }
        if (subscriberB != null && subscriberB.isConnected()) {
            subscriberB.disconnect();
            subscriberB.close();
        }

        if(mqttBrokerService !=null){
            mqttBrokerService.shutdown();
        }
    }

    /**
     * 1.连接后订阅消息，断开连接
     * 2.关闭服务
     *
     * 再同一个topic发布消息。是否能接受该topic的消息
     *
     * @throws Exception
     */
    @Test
    public void testCleanSession_maintainClientSubscriptions_withServerRestart() throws Exception {
        System.out.println("*** testCleanSession_maintainClientSubscriptions_withServerRestart ***");
        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_client.connect(options);
        m_client.subscribe("/topic/aa", 0);
        m_client.disconnect();


        mqttBrokerService.shutdown();
        mqttBrokerService = MQServer.startMqttService(new String[]{});

        // reconnect and publish
        m_client.connect(options);

        String publishStr = "Test my payload";

        m_client.publish("/topic/aa", publishStr.getBytes(UTF_8), 0, false);

        Awaitility.await().atMost(5,TimeUnit.SECONDS).until(m_messagesCollector::isMessageReceived);

        System.out.println(m_messagesCollector.retrieveMessage());

        //assertEquals(publishStr, m_messagesCollector.retrieveMessage());

        mqttBrokerService.shutdown();
    }


    /**
     * 测试客户端持久化session
     * 1.客户端上线，订阅主题a/b,session为持久化
     * 2.客户端离线，发布者发布消息到主题a/b
     * 3.客户端重新上线，接收到发布者的消息
     *
     * @throws Exception
     */
    @Test
    public void testPersistSession_maintainClientSubscriptions() throws Exception {
        System.out.println("*** testPersistSession_maintainClientSubscriptions ***");
        //客户端上线，订阅主题/a/topic,session为持久化,离线
        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_client.connect(options);
        m_client.subscribe("/a/topic", 1);
        m_client.disconnect();


        // reconnect and publish
       // m_client.connect(options);

        //客户端离线，发布者发布消息到主题/a/topic
        String publishStr = "Test my payload";
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign2.getUsername());
        options.setPassword(sign2.getPassword().toCharArray());
        m_publisher.connect(options);
        m_publisher.publish("/a/topic", publishStr.getBytes(UTF_8), 1, false);
        m_publisher.disconnect();


        //客户端重新上线，接收到发布者的消息
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_client.connect(options);
        MqttMessage mqttMessage = m_messagesCollector.retrieveMessage();

        System.out.println(mqttMessage.toString());



//        Awaitility.await().until(m_messagesCollector.retrieveMessage());
//
//        assertEquals("/a/topic", m_messagesCollector.retrieveTopic());
//        assertEquals(publishStr, m_messagesCollector.retrieveMessage());
    }

    //测试订阅消息质量为

    @Test
    public void checkSubscribersGetCorrectQosNotifications() throws Exception {
        //创建订阅者A，连接订阅a/b topic主题 消息质量1
        subscriberA = new MqttClient(HOST, sign3.getClientid(), new MqttDefaultFilePersistence());
        MessageCollector cbSubscriberA = new MessageCollector();
        subscriberA.setCallback(cbSubscriberA);

        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign3.getUsername());
        options.setPassword(sign3.getPassword().toCharArray());
        subscriberA.connect(options);
        subscriberA.subscribe("a/b", 1);

        //创建订阅者B，连接订阅a/+ topic主题 消息质量2
        subscriberB = new MqttClient(HOST, sign4.getClientid(), new MqttDefaultFilePersistence());

        MessageCollector cbSubscriberB = new MessageCollector();
        subscriberB.setCallback(cbSubscriberB);
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign4.getUsername());
        options.setPassword(sign4.getPassword().toCharArray());

        subscriberB.connect(options);
        subscriberB.subscribe("a/+", 1);


        //创建消息发布者m_client，向a/b topic主题 发布消息 消息质量2
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_client.connect(options);
        m_client.publish("a/b", "Hello world MQTT!!".getBytes(UTF_8), 1, false);
        //subscriberA订阅者受到 hello word消息质量为1
        Awaitility.await().until(cbSubscriberA::isMessageReceived);
        MqttMessage messageOnA = cbSubscriberA.retrieveMessage();
        System.out.println(new String(messageOnA.getPayload(), UTF_8));
        assertEquals("Hello world MQTT!!", new String(messageOnA.getPayload(), UTF_8));
        assertEquals(1, messageOnA.getQos());
        subscriberA.disconnect();

        //subscriberB订阅者受到 hello word消息质量为2
        Awaitility.await().until(cbSubscriberB::isMessageReceived);
        MqttMessage messageOnB = cbSubscriberB.retrieveMessage();
        assertNotNull(messageOnB, "MUST be a received message");
        assertEquals("Hello world MQTT!!", new String(messageOnB.getPayload(), UTF_8));
        assertEquals(1, messageOnB.getQos());
        subscriberB.disconnect();

    }

//    @Test
//    public void testSubcriptionDoesntStayActiveAfterARestart() throws Exception {
//        // clientForSubscribe1 connect and subscribe to /topic QoS2
//        MqttClientPersistence dsSubscriberA = new MqttDefaultFilePersistence(
//            IntegrationUtils.newFolder(tempFolder, "clientForSubscribe1").getAbsolutePath());
//
//        MqttClient clientForSubscribe1 = new MqttClient("tcp://localhost:1883", "clientForSubscribe1", dsSubscriberA);
//        MessageCollector cbSubscriber1 = new MessageCollector();
//        clientForSubscribe1.setCallback(cbSubscriber1);
//        clientForSubscribe1.connect();
//        clientForSubscribe1.subscribe("topic", 0);
//
//        // integration stop
//        m_server.stopServer();
//        System.out.println("\n\n SEVER REBOOTING \n\n");
//        // integration start
//        startServer(dbPath);
//
//        // clientForSubscribe2 connect and subscribe to /topic QoS2
//        MqttClientPersistence dsSubscriberB = new MqttDefaultFilePersistence(
//            IntegrationUtils.newFolder(tempFolder, "clientForSubscribe2").getAbsolutePath());
//        MqttClient clientForSubscribe2 = new MqttClient("tcp://localhost:1883", "clientForSubscribe2", dsSubscriberB);
//        MessageCollector cbSubscriber2 = new MessageCollector();
//        clientForSubscribe2.setCallback(cbSubscriber2);
//        clientForSubscribe2.connect();
//        clientForSubscribe2.subscribe("topic", 0);
//
//        // clientForPublish publish on /topic with QoS2 a message
//        MqttClientPersistence dsSubscriberPUB = new MqttDefaultFilePersistence(
//            IntegrationUtils.newFolder(tempFolder, "clientForPublish").getAbsolutePath());
//        MqttClient clientForPublish = new MqttClient("tcp://localhost:1883", "clientForPublish", dsSubscriberPUB);
//        clientForPublish.connect();
//        clientForPublish.publish("topic", "Hello".getBytes(UTF_8), 2, true);
//
//        // verify clientForSubscribe1 doesn't receive a notification but clientForSubscribe2 yes
//        LOG.info("Before waiting to receive 1 sec from {}", clientForSubscribe1.getClientId());
//        assertFalse(clientForSubscribe1.isConnected());
//        assertTrue(clientForSubscribe2.isConnected());
//        LOG.info("Waiting to receive 1 sec from {}", clientForSubscribe2.getClientId());
//        Awaitility.await()
//            .atMost(1, TimeUnit.SECONDS)
//            .until(cbSubscriber2::isMessageReceived);
//        assertEquals("Hello", new String(cbSubscriber2.retrieveMessage().getPayload(), UTF_8));
//    }
}
