/*
 * Copyright (c) 2012-2018 The original author or authors
 * ------------------------------------------------------
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Apache License v2.0 which accompanies this distribution.
 *
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * The Apache License v2.0 is available at
 * http://www.opensource.org/licenses/apache2.0.php
 *
 * You may elect to redistribute this code under either of these licenses.
 */

package mortals.mqtt.integration;

import com.mortals.mqtt.broker.services.MqttBrokerService;
import com.mortals.MQServer;
import com.mortals.mqtt.broker.common.util.MqttSign;
import lombok.extern.slf4j.Slf4j;
import mortals.mqtt.MessageCollector;
import org.awaitility.Awaitility;
import org.eclipse.paho.client.mqttv3.IMqttClient;
import org.eclipse.paho.client.mqttv3.MqttClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.eclipse.paho.client.mqttv3.persist.MqttDefaultFilePersistence;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;


import static java.nio.charset.StandardCharsets.UTF_8;
import static mortals.mqtt.ServerIntegrationPahoTest.HOST;
import static mortals.mqtt.ServerIntegrationPahoTest.HOST1;
import static org.junit.jupiter.api.Assertions.assertEquals;

/**
 * 消息质量测试验证
 */
@Slf4j
public class ServerIntegrationQoSValidationTest {

    IMqttClient m_subscriber;
    IMqttClient m_publisher;
    MessageCollector m_callback;
    MessageCollector m_callback_p;
    private String productKey = "cdbao";
    private String deviceName = "device1";
    private String deviceSecret = "ga7XA6KdlEeiPXQPpRbAjOZXwG8y";

    private String deviceName2 = "device2";

    private MqttSign sign;
    private MqttSign sign2;

    MqttBrokerService mqttBrokerService;

    @BeforeEach
    public void setUp() throws Exception {

        mqttBrokerService = MQServer.startMqttService(new String[]{});
        // 计算MQTT连接参数。
        sign = new MqttSign();
        sign.calculate(productKey, deviceName, deviceSecret);

        sign2 = new MqttSign();
        sign2.calculate(productKey, deviceName2, deviceSecret);

        m_subscriber = new MqttClient(HOST1, sign.getClientid(), new MemoryPersistence());
        m_callback = new MessageCollector();

        m_subscriber.setCallback(m_callback);


        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_subscriber.connect(options);

        m_publisher = new MqttClient(HOST, sign2.getClientid(), new MqttDefaultFilePersistence());

        m_callback_p = new MessageCollector();
        m_publisher.setCallback(m_callback_p);
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign2.getUsername());
        options.setPassword(sign2.getPassword().toCharArray());
        m_publisher.connect(options);
    }

    @AfterEach
    public void tearDown() throws Exception {
        if (m_publisher.isConnected()) {
            m_publisher.disconnect();
        }

        if (m_subscriber.isConnected()) {
            m_subscriber.disconnect();
        }
        mqttBrokerService.shutdown();
    }

    /**
     * 测试订阅消息质量为0，发布消息为0的情况
     *
     * @throws Exception
     */
    @Test
    public void checkSubscriberQoS0ReceiveQoS0publishes() throws Exception {
        log.info("*** checkSubscriberQoS0ReceiveQoS0publishes ***");
        m_subscriber.subscribe("/topic/bbb", 0);

        for(int i=0;i<100;i++){
            Thread.sleep(i);
            m_publisher.publish("/topic/bbb", "Hello world MQTT QoS0".getBytes(UTF_8), 0, false);
        }


        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        log.info("count:"+m_callback.getCount());
        assertEquals("Hello world MQTT QoS0", message.toString());
        assertEquals(0, message.getQos());

    }


    /**
     * 测试订阅消息为0发布消息为1的情况
     *
     * @throws Exception
     */
    @Test
    public void checkSubscriberQoS0ReceiveQoS1publishes_downgrade() throws Exception {
        log.info("*** checkSubscriberQoS0ReceiveQoS1publishes_downgrade ***");
        m_subscriber.subscribe("/topic/bbb0", 0);

        m_publisher.publish("/topic/bbb0", "Hello world MQTT QoS1".getBytes(UTF_8), 1, false);

        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS1", message.toString());
        assertEquals(0, message.getQos());
    }

    @Test
    public void checkSubscriberQoS0ReceiveQoS2publishes_downgrade() throws Exception {
        log.info("*** checkSubscriberQoS0ReceiveQoS2publishes_downgrade ***");
        m_subscriber.subscribe("/topic/bbb2", 0);

        m_publisher.publish("/topic/bbb2", "Hello world MQTT QoS2".getBytes(UTF_8), 2, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS2", message.toString());
        assertEquals(0, message.getQos());
    }

    @Test
    public void checkSubscriberQoS1ReceiveQoS0publishes() throws Exception {
        log.info("*** checkSubscriberQoS1ReceiveQoS0publishes ***");
        m_subscriber.subscribe("/topic/bbb10", 1);

        m_publisher.publish("/topic/bbb10", "Hello world MQTT QoS0".getBytes(UTF_8), 0, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS0", message.toString());
        assertEquals(0, message.getQos());
    }

    @Test
    public void checkSubscriberQoS1ReceiveQoS1publishes() throws Exception {
        log.info("*** checkSubscriberQoS1ReceiveQoS1publishes ***");
        m_subscriber.subscribe("/topic/bbb11", 1);

        m_publisher.publish("/topic/bbb11", "Hello world MQTT QoS1".getBytes(UTF_8), 1, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS1", message.toString());
        assertEquals(1, message.getQos());
    }

    @Test
    public void checkSubscriberQoS1ReceiveQoS2publishes_downgrade() throws Exception {
        log.info("*** checkSubscriberQoS1ReceiveQoS2publishes_downgrade ***");
        m_subscriber.subscribe("/topic/bbb12", 1);

        m_publisher.publish("/topic/bbb12", "Hello world MQTT QoS2".getBytes(UTF_8), 2, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS2", message.toString());
        assertEquals(1, message.getQos());
    }

    @Test
    public void checkSubscriberQoS2ReceiveQoS0publishes() throws Exception {
        log.info("*** checkSubscriberQoS2ReceiveQoS0publishes ***");
        m_subscriber.subscribe("/topic/bbb20", 2);

        m_publisher.publish("/topic/bbb20", "Hello world MQTT QoS2".getBytes(UTF_8), 0, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS2", message.toString());
        assertEquals(0, message.getQos());
    }

    @Test
    public void checkSubscriberQoS2ReceiveQoS1publishes() throws Exception {
        log.info("*** checkSubscriberQoS2ReceiveQoS1publishes ***");
        m_subscriber.subscribe("/topic", 2);

        m_publisher.publish("/topic", "Hello world MQTT QoS2".getBytes(UTF_8), 1, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS2", message.toString());
        assertEquals(1, message.getQos());
    }

    @Test
    public void checkSubscriberQoS2ReceiveQoS2publishes() throws Exception {
        log.info("*** checkSubscriberQoS2ReceiveQoS2publishes ***");
        m_subscriber.subscribe("/topic/bbb22", 2);

        m_publisher.publish("/topic/bbb22", "Hello world MQTT QoS2".getBytes(UTF_8), 2, false);
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        log.info("receive:"+message.toString());
        assertEquals("Hello world MQTT QoS2", message.toString());
        assertEquals(2, message.getQos());
    }

}
