/*
 * Copyright (c) 2012-2018 The original author or authors
 * ------------------------------------------------------
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Apache License v2.0 which accompanies this distribution.
 *
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * The Apache License v2.0 is available at
 * http://www.opensource.org/licenses/apache2.0.php
 *
 * You may elect to redistribute this code under either of these licenses.
 */

package mortals.mqtt.integration;

import com.mortals.MQServer;
import com.mortals.mqtt.broker.common.util.MqttSign;
import com.mortals.mqtt.broker.services.MqttBrokerService;
import lombok.extern.slf4j.Slf4j;
import mortals.mqtt.MessageCollector;
import org.awaitility.Awaitility;
import org.eclipse.paho.client.mqttv3.IMqttClient;
import org.eclipse.paho.client.mqttv3.MqttClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.eclipse.paho.client.mqttv3.persist.MqttDefaultFilePersistence;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.concurrent.TimeUnit;

import static java.nio.charset.StandardCharsets.UTF_8;
import static mortals.mqtt.ServerIntegrationPahoTest.HOST;
import static org.junit.jupiter.api.Assertions.assertEquals;

/**
 * 客户端集成链接测试
 */
@Slf4j
public class ServerIntegrationSessionTest {

    IMqttClient m_subscriber;
    IMqttClient m_publisher;
    MessageCollector m_callback;
    MessageCollector m_callback_p;
    private String productKey = "cdbao";
    private String deviceName = "device1";
    private String deviceSecret = "ga7XA6KdlEeiPXQPpRbAjOZXwG8y";

    private String deviceName2 = "device2";

    private MqttSign sign;
    private MqttSign sign2;

    MqttBrokerService mqttBrokerService;

    /**
     * device1 订阅者
     * device2 发布者
     * @throws Exception
     */
    @BeforeEach
    public void setUp() throws Exception {
        mqttBrokerService = MQServer.startMqttService(new String[]{});
    }

    @AfterEach
    public void tearDown() throws Exception {
        if (m_publisher.isConnected()) {
            m_publisher.disconnect();
        }

        if (m_subscriber.isConnected()) {
            m_subscriber.disconnect();
        }
        mqttBrokerService.shutdown();
    }

    /**
     * 1.设备连接鉴权。
     * 2.根据设备获取订阅信息，更新当前broker订阅树
     * 3.通知集群其它节点，删除当前设备订阅。
     * 4.针对这个topic发布消息
     * 5.订阅者能收到消息
     *
     * @throws Exception
     */
    @Test
    public void checkConnectSession() throws Exception {
        log.info("*** checkConnectSession ***");

        // 计算MQTT连接参数。
        sign = new MqttSign();
        sign.calculate(productKey, deviceName, deviceSecret);

        sign2 = new MqttSign();
        sign2.calculate(productKey, deviceName2, deviceSecret);

        m_subscriber = new MqttClient(HOST, sign.getClientid(), new MemoryPersistence());
        m_callback = new MessageCollector();

        m_subscriber.setCallback(m_callback);


        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_subscriber.connect(options);

        m_publisher = new MqttClient(HOST, sign2.getClientid(), new MqttDefaultFilePersistence());

        m_callback_p = new MessageCollector();
        m_publisher.setCallback(m_callback_p);
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign2.getUsername());
        options.setPassword(sign2.getPassword().toCharArray());
        m_publisher.connect(options);

        //打印当前订阅树

        //设备A离线打印当前订阅树
        m_subscriber.disconnect();

        //设备A再次在线打印当前订阅树
        m_subscriber.reconnect();


        TimeUnit.SECONDS.sleep(5);


    }


    /**
     * 测试离线消息
     * 1.设备连接鉴权。
     * 2.根据设备获取订阅信息，更新当前broker订阅树
     * 3.通知集群其它节点，删除当前设备订阅。
     * 4.针对这个topic发布消息
     * 5.订阅者能收到消息
     *
     * @throws Exception
     */
    @Test
    public void ClearSessionFalseToOffineMessageTest() throws Exception {
        log.info("*** ClearSessionOffineMessageTest ***");

        sign = new MqttSign();
        sign.calculate(productKey, deviceName, deviceSecret);

        sign2 = new MqttSign();
        sign2.calculate(productKey, deviceName2, deviceSecret);

        m_subscriber = new MqttClient(HOST, sign.getClientid(), new MemoryPersistence());
        m_callback = new MessageCollector();

        m_subscriber.setCallback(m_callback);


        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_subscriber.connect(options);


        m_subscriber.subscribe("/topic/p1", 1);


        m_publisher = new MqttClient(HOST, sign2.getClientid(), new MqttDefaultFilePersistence());

        m_callback_p = new MessageCollector();
        m_publisher.setCallback(m_callback_p);
        options = new MqttConnectOptions();
        options.setCleanSession(false);
        options.setKeepAliveInterval(180);
        options.setUserName(sign2.getUsername());
        options.setPassword(sign2.getPassword().toCharArray());
        m_publisher.connect(options);

        for (int i=0;i<100;i++){
            m_publisher.publish("/topic/p1", "Hello world MQTT QoS1".getBytes(UTF_8), 1, false);
            if(i==50){
                //设备A断线
                m_subscriber.disconnect();
            }

        }

        TimeUnit.SECONDS.sleep(10);
        m_subscriber.reconnect();
        TimeUnit.SECONDS.sleep(10);
        //打印当前订阅树
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        System.out.println("receive:"+message.toString());

    }


    /**
     *  clearSession true时候离线消息
     * 1.设备连接鉴权。
     * 2.根据设备获取订阅信息，更新当前broker订阅树
     * 3.通知集群其它节点，删除当前设备订阅。
     * 4.针对这个topic发布消息
     * 5.订阅者能收到消息
     *
     * @throws Exception
     */
    @Test
    public void ClearSessionTrueToOffineMessageTest() throws Exception {
        log.info("*** ClearSessionTrueToOffineMessageTest ***");

        sign = new MqttSign();
        sign.calculate(productKey, deviceName, deviceSecret);

        sign2 = new MqttSign();
        sign2.calculate(productKey, deviceName2, deviceSecret);

        m_subscriber = new MqttClient(HOST, sign.getClientid(), new MemoryPersistence());
        m_callback = new MessageCollector();

        m_subscriber.setCallback(m_callback);


        MqttConnectOptions options = new MqttConnectOptions();
        options.setCleanSession(true);
        options.setKeepAliveInterval(180);
        options.setUserName(sign.getUsername());
        options.setPassword(sign.getPassword().toCharArray());
        m_subscriber.connect(options);


        m_subscriber.subscribe("/topic/p1", 1);


        m_publisher = new MqttClient(HOST, sign2.getClientid(), new MqttDefaultFilePersistence());

        m_callback_p = new MessageCollector();
        m_publisher.setCallback(m_callback_p);
        options = new MqttConnectOptions();
        options.setCleanSession(true);
        options.setKeepAliveInterval(180);
        options.setUserName(sign2.getUsername());
        options.setPassword(sign2.getPassword().toCharArray());
        m_publisher.connect(options);

        for (int i=0;i<100;i++){
            m_publisher.publish("/topic/p1", "Hello world MQTT QoS1".getBytes(UTF_8), 1, false);
            if(i==50){
                //设备A断线
                m_subscriber.disconnect();
            }

        }

        TimeUnit.SECONDS.sleep(10);
        m_subscriber.reconnect();
        TimeUnit.SECONDS.sleep(10);
        //打印当前订阅树
        Awaitility.await().until(m_callback::isMessageReceived);
        MqttMessage message = m_callback.retrieveMessage();
        System.out.println("receive:"+message.toString());

    }

}
