/*
 * Copyright (c) 2012-2018 The original author or authors
 * ------------------------------------------------------
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Apache License v2.0 which accompanies this distribution.
 *
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * The Apache License v2.0 is available at
 * http://www.opensource.org/licenses/apache2.0.php
 *
 * You may elect to redistribute this code under either of these licenses.
 */

package mortals.mqtt.testclient;

import io.netty.bootstrap.Bootstrap;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.mqtt.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.charset.Charset;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import static io.netty.channel.ChannelFutureListener.CLOSE_ON_FAILURE;
import static io.netty.channel.ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE;

/**
 * Class used just to send and receive MQTT messages without any protocol login in action, just use
 * the encoder/decoder part.
 */
public class Client {

    public interface ICallback {

        void call(MqttMessage msg);
    }

    private static final Logger LOG = LoggerFactory.getLogger(Client.class);
    final ClientNettyMQTTHandler handler = new ClientNettyMQTTHandler();
    EventLoopGroup workerGroup;
    Channel m_channel;
    private boolean m_connectionLost;
    private ICallback callback;
    private String clientId;
    private MqttMessage receivedMsg;

    public Client(String host) {
        this(host,1883);
    }

    @SuppressWarnings("FutureReturnValueIgnored")
    public Client(String host, int port) {
        handler.setClient(this);
        workerGroup = new NioEventLoopGroup();
        try {
            Bootstrap b = new Bootstrap();
            b.group(workerGroup);
            b.channel(NioSocketChannel.class);
            b.option(ChannelOption.SO_KEEPALIVE, true);
            b.handler(new ChannelInitializer<SocketChannel>() {

                @Override
                public void initChannel(SocketChannel ch) throws Exception {
                    ChannelPipeline pipeline = ch.pipeline();
                    pipeline.addLast("decoder", new MqttDecoder());
                    pipeline.addLast("encoder", MqttEncoder.INSTANCE);
                    pipeline.addLast("handler", handler);
                }
            });

            // Start the client.
            m_channel = b.connect(host, port).sync().channel();
        } catch (Exception ex) {
            LOG.error("Error received in client setup", ex);
            workerGroup.shutdownGracefully();
        }
    }

    public Client clientId(String clientId) {
        this.clientId = clientId;
        return this;
    }

    public void connect(String willTestamentTopic, String willTestamentMsg) {
        MqttFixedHeader mqttFixedHeader = new MqttFixedHeader(
                MqttMessageType.CONNECT,
                false,
                MqttQoS.AT_MOST_ONCE,
                false,
                0);
        MqttConnectVariableHeader mqttConnectVariableHeader = new MqttConnectVariableHeader(
                MqttVersion.MQTT_3_1.protocolName(),
                MqttVersion.MQTT_3_1.protocolLevel(),
                false,
                false,
                false,
                MqttQoS.AT_MOST_ONCE.value(),
                true,
                true,
                2);
        MqttConnectPayload mqttConnectPayload = new MqttConnectPayload(
                this.clientId,
                willTestamentTopic,
                willTestamentMsg.getBytes(Charset.forName("UTF-8")),
                null,
                null);
        MqttConnectMessage connectMessage = new MqttConnectMessage(
                mqttFixedHeader,
                mqttConnectVariableHeader,
                mqttConnectPayload);

        /*
         * ConnectMessage connectMessage = new ConnectMessage();
         * connectMessage.setProtocolVersion((byte) 3); connectMessage.setClientID(this.clientId);
         * connectMessage.setKeepAlive(2); //secs connectMessage.setWillFlag(true);
         * connectMessage.setWillMessage(willTestamentMsg.getBytes());
         * connectMessage.setWillTopic(willTestamentTopic);
         * connectMessage.setWillQos(MqttQoS.AT_MOST_ONCE.byteValue());
         */

        doConnect(connectMessage);
    }

    public void connect() {
        MqttConnectMessage connectMessage = MqttMessageBuilders.connect().protocolVersion(MqttVersion.MQTT_3_1_1)
                .clientId(clientId).password("1234546").keepAlive(2) // secs
                .willFlag(false).willQoS(MqttQoS.AT_MOST_ONCE).build();

        doConnect(connectMessage);
    }

    public void doConnect(MqttConnectMessage connectMessage) {
        final CountDownLatch latch = new CountDownLatch(1);
        this.setCallback(msg -> {
            receivedMsg = msg;
            latch.countDown();
        });

        this.sendMessage(connectMessage);

        try {
            latch.await(1000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            throw new RuntimeException("Cannot receive message in 200 ms", e);
        }
        if (!(this.receivedMsg instanceof MqttConnAckMessage)) {
            MqttMessageType messageType = this.receivedMsg.fixedHeader().messageType();
            throw new RuntimeException("Expected a CONN_ACK message but received " + messageType);
        }
    }

    public void setCallback(ICallback callback) {
        this.callback = callback;
    }

    public void sendMessage(MqttMessage msg) {
        m_channel.writeAndFlush(msg).addListener(FIRE_EXCEPTION_ON_FAILURE);
    }

    public MqttMessage lastReceivedMessage() {
        return this.receivedMsg;
    }

    void messageReceived(MqttMessage msg) {
        LOG.info("Received message {}", msg);
        if (this.callback != null) {
            this.callback.call(msg);
        }
    }

    void setConnectionLost(boolean status) {
        m_connectionLost = status;
    }

    public boolean isConnectionLost() {
        return m_connectionLost;
    }

    @SuppressWarnings("FutureReturnValueIgnored")
    public void close() throws InterruptedException {
        // Wait until the connection is closed.
        m_channel.closeFuture().sync().addListener(CLOSE_ON_FAILURE);
        if (workerGroup == null) {
            throw new IllegalStateException("Invoked close on an Acceptor that wasn't initialized");
        }
        workerGroup.shutdownGracefully();
    }
}
