package com.mortals.xhx.module.product.web;

import com.alibaba.fastjson.JSONObject;
import com.mortals.framework.exception.AppException;
import com.mortals.framework.model.DbOrderKind;
import com.mortals.framework.service.ICacheService;
import com.mortals.xhx.base.system.param.service.ParamService;
import com.mortals.xhx.common.code.*;
import com.mortals.xhx.module.customer.model.CustomerEntity;
import com.mortals.xhx.module.customer.service.CustomerService;
import com.mortals.xhx.module.device.model.DeviceEntity;
import com.mortals.xhx.module.device.model.DeviceQuery;
import com.mortals.xhx.module.device.service.DeviceService;
import com.mortals.xhx.module.device.service.DeviceTopicService;
import com.mortals.xhx.module.product.model.ProductFunctionItemEntity;
import com.mortals.xhx.module.product.model.ProductFunctionItemQuery;
import com.mortals.xhx.module.product.service.ProductFunctionItemService;
import com.mortals.xhx.service.common.code.DataTypeEnum;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import com.mortals.framework.model.Context;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.mortals.framework.web.BaseCRUDJsonMappingController;
import com.mortals.xhx.module.product.model.ProductEntity;
import com.mortals.xhx.module.product.service.ProductService;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * ProductController
 * <p>
 * 产品 控制器
 *
 * @author zxfei
 * @date 2021-11-09
 */
@RestController
@RequestMapping("product")
public class ProductController extends BaseCRUDJsonMappingController<ProductService, ProductForm, ProductEntity, Long> {

    @Autowired
    private CustomerService customerService;

    @Autowired
    private ProductFunctionItemService productFunctionItemService;

    @Autowired
    private DeviceService deviceService;

    @Autowired
    private DeviceTopicService deviceTopicService;

    @Autowired
    private ICacheService cacheService;

    public ProductController() {
        super.setFormClass(ProductForm.class);
        super.setModuleDesc("产品表");
    }

    @Override
    protected void init(HttpServletRequest request, HttpServletResponse response, ProductForm form,
                        Map<String, Object> model, Context context) {
        this.addDict(model, "productType", ProductTypeEnum.getEnumMap());
        this.addDict(model, "nodeType", NodeTypeEnum.getEnumMap());
        this.addDict(model, "connGateProtocol", ConnGateProtocolEnum.getEnumMap());
        this.addDict(model, "networkType", NetworkTypeEnum.getEnumMap());
        this.addDict(model, "dataType", DataTypeEnum.getEnumMap());
        this.addDict(model, "checkType", CheckTypeEnum.getEnumMap());
        this.addDict(model, "authType", AuthTypeEnum.getEnumMap());
        this.addDict(model, "productStatus", ProductStatusEnum.getEnumMap());
        this.addDict(model, "qos", QosTypeEnum.getEnumMap());
        this.addDict(model, "customerId",
                customerService.getCacheList().stream().collect(Collectors.toMap(e -> e.getId().toString(), CustomerEntity::getLoginName)));
        super.init(request, response, form, model, context);
    }

    @Override
    protected void doListBefore(HttpServletRequest request, HttpServletResponse response, ProductForm form, Map<String, Object> model, Context context) throws AppException {
        Map<String, String> orderCols = new HashMap<>();
        orderCols.put("id", DbOrderKind.DESC.name());
        form.getQuery().setOrderCols(orderCols);
        super.doListBefore(request, response, form, model, context);
    }

    /**
     * 修改产品状态
     * 发布/停止发布（开发中）
     */
    @PostMapping(value = "editProductStatus")
    public String editProductStatus(@RequestParam(value = "id", required = false) Long id, @RequestParam(value = "status",
            required = false) Integer status) {
        JSONObject jsonObject = new JSONObject();
        jsonObject.put(KEY_RESULT_CODE, VALUE_RESULT_SUCCESS);
        try {
            if (ProductStatusEnum.DEVELOP.getValue() == status) {//停止发布（开发中）
                //TODO 此时需不需要更新产品下设备topic更新redis？
                updateProductStatus(id, status);
                jsonObject.put(KEY_RESULT_MSG, "撤销发布成功");
                return jsonObject.toJSONString();
            }
            //发布
            ProductFunctionItemQuery query = new ProductFunctionItemQuery();
            query.setProductId(id);
            List<ProductFunctionItemEntity> itemList = productFunctionItemService.find(query, new Context());
            if (null != itemList && itemList.size() != 0) {
                for (ProductFunctionItemEntity itemEntity : itemList) {
                    if (itemEntity.getProductFunctionStatus() != ProductFunctionStatusEnum.USER.getValue()) {
                        jsonObject.put(KEY_RESULT_CODE, VALUE_RESULT_FAILURE);
                        jsonObject.put(KEY_RESULT_MSG, "产品下物模型节点未发布");
                        return jsonObject.toJSONString();
                    }
                }
            } else {
                jsonObject.put(KEY_RESULT_CODE, VALUE_RESULT_FAILURE);
                jsonObject.put(KEY_RESULT_MSG, "产品下无物模型");
                return jsonObject.toJSONString();
            }
            //更新产品下设备topic与缓存对应关系
            updateDeviceTopic(id);
            updateProductStatus(id, status);
            jsonObject.put(KEY_RESULT_MSG, "发布成功");
            return jsonObject.toJSONString();
        } catch (Exception e) {
            log.error("发布/停止发布出现异常:" + e.toString());
            jsonObject.put(KEY_RESULT_CODE, VALUE_RESULT_FAILURE);
            jsonObject.put(KEY_RESULT_MSG, "发布/停止发布出现异常");
            return jsonObject.toJSONString();
        }
    }

    /**
     * 根据产品id更新产品状态
     *
     * @param id
     * @param status
     */
    public void updateProductStatus(Long id, Integer status) {
        ProductEntity entity = new ProductEntity();
        entity.setId(id);
        entity.setProductStatus(status);
        service.update(entity, new Context());
    }

    /**
     * 更新产品下设备topic与缓存对应关系
     *
     * @param id
     */
    public void updateDeviceTopic(Long id) throws AppException {
        DeviceQuery dQuery = new DeviceQuery();
        dQuery.setProductId(id);
        List<DeviceEntity> dList = deviceService.find(dQuery, new Context());
        //先删除以前的设备topic与缓存关系
        if (null != dList && dList.size() != 0) {
            Long[] dtIds = dList.stream().map(DeviceEntity::getId).toArray(Long[]::new);
            deviceService.delDeviceTopicByDevIds(dtIds, new Context());
        } else {//产品下还没有设备
            return;
        }
        //再新增最新的设备topic与缓存关系
        dList.forEach((dEntity) -> {
            deviceService.updateDeviceTopic(dEntity, new Context(), false);
        });
    }
}